# Copyright 2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from attrs import define, field
import re
from rich.progress import track
from slugify import slugify
from typing import Iterable, TYPE_CHECKING

from .. import constants
from ..config import primary_console
from ..containers import SlottedDict
from ..exceptions import AnthologyDuplicateIDError
from ..text import StopWords
from ..utils.logging import get_logger
from .paper import Paper

if TYPE_CHECKING:
    from .collection import Collection
    from .index import CollectionIndex


log = get_logger()


BIBKEY_MAX_NAMES = 2
"""The maximum number of names to consider when generating bibkeys."""


@define
class BibkeyIndex(SlottedDict[Paper]):
    """Index object which collects citation keys for all papers.

    Provides dictionary-like functionality mapping bibkeys to [Paper][acl_anthology.collections.collection.Paper] objects in the Anthology.

    Attributes:
        parent: The parent CollectionIndex instance to which this index belongs.
        is_data_loaded: A flag indicating whether the XML directory has already been indexed.
    """

    parent: CollectionIndex = field(repr=False, eq=False)
    is_data_loaded: bool = field(init=False, repr=True, default=False)

    def generate_bibkey(self, paper: Paper) -> str:
        """Generate a unique bibkey for the given paper.

        Parameters:
            paper: The paper for which a bibkey should be generated.

        Returns:
            The generated bibkey.

        Note:
            Calling this function will _not change_ the paper's bibkey nor add the bibkey to the index.
        """
        if not self.is_data_loaded:
            self.load()

        title_words = None

        if paper.is_frontmatter:
            # Proceedings volumes/frontmatter use {venue}-{year}-{volume_id}
            bibkey = (
                f"{slugify(paper.parent.venue_acronym)}-{paper.year}-{paper.volume_id}"
            )
        else:
            # Generate slugified author string, using '-etal' if necessary
            namespecs = paper.authors if paper.authors else paper.get_editors()
            if not namespecs:
                bibnames = "nn"
            elif len(namespecs) > BIBKEY_MAX_NAMES:
                bibnames = f"{slugify(namespecs[0].last)}-etal"
            else:
                bibnames = "-".join(slugify(ns.last) for ns in namespecs)

            # Generate slugified and filtered list of title words
            title_words = [
                word
                for word in slugify(paper.title.as_text()).split("-")
                if not StopWords.contains(word)
            ]
            # Edge case: ensure title_words is never empty
            if not title_words:
                title_words = [
                    word for word in slugify(paper.title.as_text()).split("-")
                ] + ["0"]

            # Regular papers use {authors}-{year}-{first_title_words}
            bibkey = f"{bibnames}-{paper.year}-{title_words.pop(0)}"

        # Guarantee uniqueness
        while bibkey in self.data:
            if title_words:
                # If we have unused title words, take the next one
                bibkey = f"{bibkey}-{title_words.pop(0)}"
            else:
                # Otherwise, add a number, starting from 2
                if (m := re.search(r"-([0-9]+)$", bibkey)) is not None:
                    number = int(m.group(1)) + 1
                    bibkey = m.re.sub(f"-{number}", bibkey)
                else:
                    bibkey = f"{bibkey}-2"

        return bibkey

    def _index_paper(self, bibkey: str, paper: Paper) -> str:
        """Add a paper to the index.

        If the supplied bibkey is [`constants.NO_BIBKEY`][acl_anthology.constants.NO_BIBKEY], this will trigger creation of a new bibkey.

        Warning:
            This function should not be called manually. It is invoked automatically when a paper's bibkey is changed.

        Returns:
            The bibkey that was indexed.

        Raises:
            AnthologyDuplicateIDError: If the bibkey is not [`constants.NO_BIBKEY`][acl_anthology.constants.NO_BIBKEY] and is already in the index pointing to another paper.
        """
        if not self.is_data_loaded:
            self.load()

        if paper.bibkey in self.data:
            del self.data[paper.bibkey]

        if bibkey == constants.NO_BIBKEY:
            bibkey = self.generate_bibkey(paper)
        elif bibkey in self.data:
            raise AnthologyDuplicateIDError(
                bibkey,
                f"Cannot index bibkey '{bibkey}' for paper {paper.full_id}; already assigned to {self.data[bibkey].full_id}",
            )

        self.data[bibkey] = paper
        return bibkey

    def load(self) -> None:
        """Load an index of bibkeys."""
        # This function exists so we can later add the option to read the index
        # from a cache if it doesn't need re-building.
        if self.is_data_loaded:
            return
        self.build(show_progress=self.parent.parent.verbose)
        self.is_data_loaded = True

    def reset(self) -> None:
        """Reset the index."""
        self.data = {}
        self.is_data_loaded = False

    def build(self, show_progress: bool = False) -> None:
        """Load the entire Anthology data and build an index of bibkeys.

        Raises:
            AnthologyDuplicateIDError: If a non-unique bibkey is encountered.  In case of multiple errors, only one exception is raised at the end, and all errors are sent to the logger.
        """
        self.reset()
        # Go through every single paper
        if not show_progress:
            iterator: Iterable[Collection] = self.parent.values()
        else:
            iterator = track(
                self.parent.values(),
                total=len(self.parent),
                description="Building bibkey index...",
                console=primary_console,
            )
        errors = []
        for collection in iterator:
            for paper in collection.papers():
                if paper.bibkey in self.data:
                    log.error(
                        f"Paper {paper.full_id} has bibkey {paper.bibkey}, which is already assigned to paper {self.data[paper.bibkey].full_id}"
                    )
                    errors.append(paper.bibkey)
                self.data[paper.bibkey] = paper
        if errors:
            raise AnthologyDuplicateIDError(
                errors, "There were duplicate bibkeys while building the bibkey index."
            )
        self.is_data_loaded = True



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for representing and resolving file references."""

from __future__ import annotations

import sys
from attrs import define, field, validators as v
from lxml import etree
from lxml.builder import E
from pathlib import Path
from typing import cast, ClassVar, Optional, TYPE_CHECKING
from zlib import crc32

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    from _typeshed import StrPath

from .config import config
from .utils.xml import xsd_boolean


def compute_checksum(value: bytes) -> str:
    """Compute the checksum of a byte string.

    Parameters:
        value: Any byte string.

    Returns:
        The checksum of the byte string as an eight-character, hex-formatted string.
    """
    checksum = crc32(value) & 0xFFFFFFFF
    return f"{checksum:08x}"


def compute_checksum_from_file(path: StrPath) -> str:
    """Compute the checksum of a file.

    Parameters:
        path: The path to a file.

    Returns:
        The checksum of the file's contents.
    """
    with open(path, "rb") as f:
        return compute_checksum(f.read())


@define
class FileReference:
    """Base class for all references to local or remote files in the XML data.

    Do not instantiate directly; use the sub-classes instead.

    Attributes:
        template_field (str): The URL formatting template to use.  Set by the sub-classes.
        name (str): The file reference (as found in the XML), typically a URL or an internal filename.
        checksum (Optional[str]): The CRC32 checksum for the file.  Only specified for internal filenames.
    """

    name: str = field(converter=str)
    checksum: Optional[str] = field(
        default=None, validator=v.optional(v.instance_of(str))
    )
    template_field: ClassVar[str] = ""

    @property
    def is_local(self) -> bool:
        """Whether this is a local filename."""
        return "://" not in self.name

    @property
    def url(self) -> str:
        """The URL at which this file can be accessed."""
        if "://" in self.name:
            return self.name
        return cast(str, config[self.template_field]).format(self.name)

    @classmethod
    def from_file(cls, filename: StrPath) -> Self:
        """Instantiate a new file reference from a file.

        This automatically computes the checksum for the file and determines its name.

        The name of the returned reference will depend on the configured template string; for example, if this function is called on the [PDFReference][acl_anthology.files.PDFReference] class and given a filename ending in `".pdf"`, and [`config.pdf_location_template`][acl_anthology.config.DefaultConfig] ends in `".pdf"`, this means the filename should and will be stored _without_ the `".pdf"` suffix.

        Parameters:
            filename: The path to the file.

        Returns:
            A file reference for the given file.

        Raises:
            FileNotFoundError: If filename does not point to an existing file.
        """
        if not (path := Path(filename)).is_file():
            raise FileNotFoundError(f"Not a file: {filename}")

        if cast(str, config[cls.template_field]).endswith(path.suffix):
            name = path.stem
        else:
            name = path.name

        checksum = compute_checksum_from_file(path)
        return cls(name=name, checksum=checksum)

    @classmethod
    def from_xml(cls, elem: etree._Element) -> Self:
        """Instantiate a new file reference from a corresponding XML element."""
        checksum = elem.get("hash")
        return cls(name=str(elem.text), checksum=str(checksum) if checksum else None)

    def to_xml(self, tag: str = "url") -> etree._Element:
        """
        Arguments:
            tag: Name of outer tag in which this file reference should be wrapped.  Defaults to "url".

        Returns:
            A serialization of this file reference in Anthology XML format.
        """
        elem = etree.Element(tag)
        elem.text = self.name
        if self.is_local and self.checksum is not None:
            elem.set("hash", str(self.checksum))
        return elem


@define
class PDFReference(FileReference):
    """Reference to a PDF file."""

    template_field: ClassVar[str] = "pdf_location_template"


@define
class PDFThumbnailReference(FileReference):
    """Reference to a PDF thumbnail image."""

    template_field: ClassVar[str] = "pdf_thumbnail_location_template"


@define
class AttachmentReference(FileReference):
    """Reference to an attachment."""

    # TODO: attachments must be local files according to the schema

    template_field: ClassVar[str] = "attachment_location_template"


@define
class EventFileReference(FileReference):
    """Reference to an event-related file."""

    template_field: ClassVar[str] = "event_location_template"


@define
class VideoReference(FileReference):
    """Reference to a video."""

    template_field: ClassVar[str] = "video_location_template"
    permission: bool = field(default=True, converter=bool)

    @classmethod
    def from_xml(cls, elem: etree._Element) -> Self:
        name = str(elem.get("href"))
        if (value := elem.get("permission")) is not None:
            return cls(name=name, permission=xsd_boolean(str(value)))
        else:
            return cls(name=name)

    def to_xml(self, tag: str = "video") -> etree._Element:
        elem = E.video(href=self.name)
        if not self.permission:
            elem.set("permission", "false")
        return elem



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for simpler string manipulation."""

from typing import cast, Optional
import unicodedata

_MONTH_TO_NUM: dict[str, int] = {
    "january": 1,
    "february": 2,
    "march": 3,
    "april": 4,
    "may": 5,
    "june": 6,
    "july": 7,
    "august": 8,
    "september": 9,
    "october": 10,
    "november": 11,
    "december": 12,
}

_CLEAN_UNICODE_TRANS = str.maketrans(
    {
        "\u00ad": "",  # soft hyphen
        "\u2010": "-",  # hyphen
        "\u2011": "-",  # non-breaking hyphen
    }
)


def clean_unicode(s: str) -> str:
    """Performs some opinionated string normalization.

    Originally implemented in <https://github.com/acl-org/acl-anthology/blob/master/bin/normalize_anth.py> by David Wei Chiang, this is intended to standardize how we represent certain Unicode characters in our data, e.g. by decomposing ligatures, removing "invisible" soft hyphens, etc.

    Arguments:
        s: Any text string.

    Returns:
        The cleaned up string.
    """
    s = s.translate(_CLEAN_UNICODE_TRANS)

    # Some sources encode an i with an accent above using dotless i,
    # which must be converted to normal i
    start = 0
    while (idx := s.find("ı", start)) > -1:
        # bug: we should only be looking for accents above, not below
        if unicodedata.category(s[idx + 1]) == "Mn":
            s = f"{s[:idx]}i{s[idx+1:]}"
        start = idx + 1

    # Selectively apply compatibility decomposition.
    # This converts, e.g., ﬁ to fi and ： to :, but not ² to 2.
    # Unsure: … to ...
    # More classes could be added here.
    def decompose(c: str) -> str:
        d = unicodedata.decomposition(c)
        if d and d.split(None, 1)[0] in ["<compat>", "<wide>", "<narrow>", "<noBreak>"]:
            return unicodedata.normalize("NFKD", c)
        else:
            return c

    s = "".join(map(decompose, s))

    # Convert combining characters when possible
    s = unicodedata.normalize("NFC", s)

    return s


def interpret_pages(text: str) -> tuple[str, str]:
    """Splits up a 'pages' field into first and last page.

    Arguments:
        text: A text string representing a page range.

    Returns:
        A tuple `(first_page, last_page)`; if a known separator was found, this is the result of splitting the input on the separator; otherwise, we assume that the field contains a single page.
    """
    for s in ("--", "-", "–"):
        if text.count(s) == 1:
            return cast(tuple[str, str], tuple(text.split(s)))
    return (text, text)


def month_str2num(text: str) -> Optional[int]:
    """Convert a month string to a number, e.g. February -> 2

    Arguments:
        text: A text string representing a month value.

    Returns:
        None if the string doesn't correspond to a month; the numeric month value otherwise.

    Note:
        We're not using Python's datetime here since its behaviour depends on the system locale.
    """
    return _MONTH_TO_NUM.get(text.lower(), None)


def remove_extra_whitespace(text: str) -> str:
    """
    Arguments:
        text: An arbitrary string.

    Returns:
        The input string without newlines and consecutive whitespace replaced by a single whitespace character.
    """
    text = text.replace("\n", "").strip()
    # This was profiled to be 2x-4x faster than using re.sub();
    # also cf. https://stackoverflow.com/a/15913564
    while "  " in text:
        text = text.replace("  ", " ")
    return text



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from attrs import define, field, validators as v
from functools import cache, cached_property
from lxml import etree
from lxml.builder import E
import re
from slugify import slugify
from typing import Any, Optional, cast, TypeAlias
import yaml

try:
    from yaml import CDumper as Dumper
except ImportError:  # pragma: no cover
    from yaml import Dumper  # type: ignore

from ..utils.latex import latex_encode


@define(frozen=True)
class Name:
    """A person's name.

    Note:
        Name objects are _frozen_, meaning they are immutable.  This allows them to be used as dictionary keys, but means that in order to change a name somewhere, you need to replace it with a new `Name` instance.

    Attributes:
        first: First name part. Can be given as `None` for people who
            only have a single name, but cannot be omitted.
        last: Last name part.
        script: The script in which the name is written; only used for non-Latin script name variants.

    Examples:
        >>> Name("Yang", "Liu")
        >>> Name(last="Liu", first="Yang")
        >>> Name(None, "Mausam")
    """

    first: Optional[str] = field(
        eq=lambda x: x if x else None, validator=v.optional(v.instance_of(str))
    )
    last: str = field(validator=v.instance_of(str))
    script: Optional[str] = field(
        default=None, repr=False, eq=False, validator=v.optional(v.instance_of(str))
    )

    def as_first_last(self) -> str:
        """
        Returns:
            The person's full name in the form '{first} {last}'.
        """
        if not self.first:
            return self.last
        return f"{self.first} {self.last}"

    def as_last_first(self) -> str:
        """
        Returns:
            The person's full name in the form '{last}, {first}'.
        """
        if not self.first:
            return self.last
        return f"{self.last}, {self.first}"

    def as_full(self) -> str:
        """
        Builds the full name, determining the appropriate format based on the script.

        Returns:
            For Han names, this will be '{last}{first}'; for other scripts (or if no script is given), this will be '{first} {last}'.
        """
        if not self.first:
            return self.last
        if self.script == "hani":
            return f"{self.last}{self.first}"
        return f"{self.first} {self.last}"

    @cache
    def as_bibtex(self) -> str:
        """
        Returns:
            The person's full name as formatted in a BibTeX entry.
        """
        return latex_encode(self.as_last_first())

    def score(self) -> float:
        """
        Returns:
            A score for this name that is intended for comparing different names that generate the same ID.  Names that are more likely to be the correct canonical variant should return higher scores via this function.
        """
        name = self.as_first_last()
        # Prefer longer variants
        score = float(len(name))
        # Prefer variants with non-ASCII characters or dashes
        score += sum((ord(c) > 127 or c == "-") for c in name)
        # Penalize upper-case characters after word boundaries
        score -= sum(any(c.isupper() for c in w[1:]) for w in re.split(r"\W+", name))
        # Penalize lower-case characters at word boundaries
        score -= sum(w[0].islower() if w else 0 for w in re.split(r"\W+", name))
        if name[0].islower():  # extra penalty for first name
            score -= 1
        # Penalize first names that are longer than last names (this is
        # intended to make a difference when a person has both "C, A B" and "B
        # C, A" as names)
        if self.first and len(self.first) > len(self.last):
            score += 0.5
        return score

    @cache
    def slugify(self) -> str:
        """
        Returns:
            A [slugified string](https://github.com/un33k/python-slugify#how-to-use) of the full name.
        """
        return slugify(self.as_first_last())

    @classmethod
    def from_dict(cls, name: dict[str, str]) -> Name:
        """
        Parameters:
            name: A dictionary with "first" and "last" keys.

        Returns:
            A corresponding Name object.
        """
        return cls(
            name.get("first"),
            name["last"],
        )

    @classmethod
    def from_xml(cls, variant: etree._Element) -> Name:
        """
        Parameters:
            variant: An XML element of a `<variant>` block.

        Returns:
            A corresponding Name object.

        Note:
            This will work for `<author>` and `<editor>` tags as well, but those
            are more efficiently parsed within
            [NameSpecification.from_xml()][acl_anthology.people.name.NameSpecification.from_xml].
        """
        first: Optional[str] = None
        last: Optional[str] = None
        script = variant.get("script")

        for element in variant:
            if element.tag == "first":
                first = element.text
            elif element.tag == "last":
                last = element.text
        return cls(first, cast(str, last), script)

    @classmethod
    def from_string(cls, name: str) -> Name:
        """Instantiate a Name from a single string.

        Parameters:
            name: A name string given as either "{first} {last}" or "{last}, {first}".

        Returns:
            A corresponding Name object.

        Raises:
            ValueError: If `name` cannot be unambiguously parsed into first/last components; in this case, you should instantiate Name directly instead.
        """
        name = name.strip()
        if ", " in name:
            components = name.split(", ")[::-1]
        else:
            components = name.split(" ")
        if len(components) == 1:
            return cls(None, components[0])
        elif len(components) > 2:
            raise ValueError(
                f"Name string cannot be unambiguously parsed into first/last components: {name}"
            )
        return cls(components[0], components[1])

    @classmethod
    def from_(cls, name: ConvertableIntoName) -> Name:
        """Instantiate a Name dynamically from any type that can be converted into a Name.

        Parameters:
            name: A name as a string, dict, tuple, or Name instance.

        Returns:
            A corresponding Name object.

        Raises:
            ValueError:
            TypeError:
        """
        if isinstance(name, cls):
            return name
        elif isinstance(name, dict):
            return cls.from_dict(name)
        elif isinstance(name, tuple):
            return cls(*name)
        elif isinstance(name, str):
            return cls.from_string(name)
        else:  # pragma: no cover
            raise TypeError(f"Cannot instantiate Name from {type(name)}")

    def to_xml(self, tag: str = "variant") -> etree._Element:
        """
        Arguments:
            tag: Name of outer tag in which the name should be wrapped.

        Returns:
            A serialization of this name in Anthology XML format.
        """
        elem = etree.Element(tag)
        elem.extend(
            (
                E.first(self.first) if self.first else E.first(),
                E.last(self.last),
            )
        )
        if self.script is not None:
            elem.set("script", self.script)
        return elem


ConvertableIntoName: TypeAlias = Name | str | tuple[Optional[str], str] | dict[str, str]
"""A type that can be converted into a Name instance."""


def _Name_from(value: Any) -> Name:
    return Name.from_(value)


@define
class NameSpecification:
    """A name specification on a paper etc., containing additional data fields for information or disambiguation besides just the name.

    Attributes:
        name: The person's name.
        id: Unique ID for the person that this name refers to.
        orcid: An ORCID that was supplied together with this name.
        affiliation: Professional affiliation.
        variants: Variant spellings of this name in different scripts.

    Note:
        The `variants` attribute is only intended for name variants stored via the
        `<variant>` tag in the XML, i.e., for a name that has a variant in a different
        script.  It is _not_ used when an author has published under different names
        (for this functionality, see [Person][acl_anthology.people.person.Person]).
    """

    name: Name = field(converter=_Name_from)
    id: Optional[str] = field(default=None, validator=v.optional(v.instance_of(str)))
    orcid: Optional[str] = field(default=None, validator=v.optional(v.instance_of(str)))
    affiliation: Optional[str] = field(
        default=None, validator=v.optional(v.instance_of(str))
    )
    variants: list[Name] = field(
        factory=list,
        validator=v.deep_iterable(
            member_validator=v.instance_of(Name),
            iterable_validator=v.instance_of(list),
        ),
    )

    def __hash__(self) -> int:
        return hash((self.name, self.id, self.affiliation, tuple(self.variants)))

    @property
    def first(self) -> Optional[str]:
        """The first name component."""
        return self.name.first

    @property
    def last(self) -> str:
        """The last name component."""
        return self.name.last

    @cached_property
    def citeproc_dict(self) -> dict[str, str]:
        """A citation object corresponding to this name for use with CiteProcJSON."""
        if not self.name.first:
            return {"family": self.name.last}
        return {"family": self.name.last, "given": self.name.first}

    @classmethod
    def from_xml(cls, person: etree._Element) -> NameSpecification:
        """
        Parameters:
            person: An XML element of an `<author>` or `<editor>` block.

        Returns:
            A corresponding NameSpecification object.
        """
        first: Optional[str] = None
        last: Optional[str] = None
        affiliation: Optional[str] = None
        variants = []

        for element in person:
            if element.tag == "first":
                first = element.text
            elif element.tag == "last":
                last = element.text
            elif element.tag == "affiliation":
                affiliation = element.text
            elif element.tag == "variant":
                variants.append(Name.from_xml(element))

        return cls(
            Name(first, cast(str, last)),
            id=person.get("id"),
            orcid=person.get("orcid"),
            affiliation=affiliation,
            variants=variants,
        )

    def to_xml(self, tag: str = "author") -> etree._Element:
        """
        Arguments:
            tag: Name of outer tag in which the name should be wrapped.

        Returns:
            A serialization of this name in Anthology XML format.
        """
        elem = etree.Element(tag)
        if self.id is not None:
            elem.set("id", self.id)
        if self.orcid is not None:
            elem.set("orcid", self.orcid)
        elem.extend(
            (
                E.first(self.first) if self.first else E.first(),
                E.last(self.last),
            )
        )
        if self.affiliation is not None:
            elem.append(E.affiliation(self.affiliation))
        for variant in self.variants:
            elem.append(variant.to_xml())
        return elem


class _YAMLName(yaml.YAMLObject):
    """YAMLObject representing names.

    This exists to serialize names in "flow" style (i.e. one-liner `{first: ..., last: ...}`) without having to force flow style on the entire YAML document.
    """

    yaml_dumper = Dumper
    yaml_tag = "tag:yaml.org,2002:map"  # serialize like a dictionary
    yaml_flow_style = True  # force flow style

    def __init__(self, name: Name) -> None:
        if name.first is not None:
            self.first = name.first
        self.last = name.last
        if name.script is not None:
            self.script = name.script



# Copyright 2023-2024 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import attrs
from attrs import define, field, setters
from enum import Enum
from typing import Any, Iterator, Optional, Sequence, TYPE_CHECKING
from ..exceptions import AnthologyException, AnthologyInvalidIDError
from ..utils.attrs import auto_validate_types
from ..utils.ids import (
    AnthologyIDTuple,
    build_id_from_tuple,
    is_valid_orcid,
    is_verified_person_id,
)
from . import Name

if TYPE_CHECKING:
    from . import NameSpecification
    from ..anthology import Anthology
    from ..collections import Paper, Volume


class NameLink(Enum):
    """How a Name was connected to a Person."""

    EXPLICIT = "explicit"
    """Name is explicitly listed in `people.yaml` file."""

    INFERRED = "inferred"
    """Name was connected to this Person via slug matching heuristic."""


def _name_list_converter(
    name_list: Sequence[Name | tuple[Name, NameLink]],
) -> list[tuple[Name, NameLink]]:
    return [
        (item, NameLink.EXPLICIT) if isinstance(item, Name) else item
        for item in name_list
    ]


def _update_person_index(person: Person, attr: attrs.Attribute[Any], value: str) -> str:
    """Update the [PersonIndex][acl_anthology.people.index.PersonIndex].

    Intended to be called from `on_setattr` of an [attrs.field][].
    """
    index = person.parent.people
    if attr.name == "id":
        index._update_id(person.id, value)
    elif attr.name == "orcid":
        index._update_orcid(person.id, person.orcid, value)
    return value


@define(field_transformer=auto_validate_types)
class Person:
    """A natural person.

    Info:
        The connection between persons and Anthology items is derived from [name specifications][acl_anthology.people.name.NameSpecification] on volumes and papers, and not stored explicitly. This means that Person objects **cannot be used to make changes to paper metadata**, e.g. which person a paper is associated with or under which name; change the information on papers instead.

        Person objects **can** be used to make changes to metadata that appears in `people.yaml`, such as ORCID, comment, degree, and alternative names for this person.

    Attributes:
        id: A unique ID for this person.  Do not change this attribute directly; use [`change_id()`][acl_anthology.people.person.Person.change_id], [`make_explicit()`][acl_anthology.people.person.Person.make_explicit], or [`merge_with_explicit()`][acl_anthology.people.person.Person.merge_with_explicit] instead.
        parent: The parent Anthology instance to which this person belongs.
        item_ids: A list of volume and/or paper IDs this person has authored or edited.
        orcid: The person's ORCID.
        comment: A comment for disambiguation purposes.
        degree: The person's institution of highest degree, for disambiguation purposes.
        similar_ids: A list of person IDs with names that should be considered similar to this one.  Do **not** use this to _find_ people with similar names; that should be done via [`PersonIndex.similar`][acl_anthology.people.index.PersonIndex].  This attribute can be used to explicitly add more "similar IDs" that are not automatically derived via similar names.
        disable_name_matching: If True, no items should be assigned to this person unless they explicitly specify this person's ID.
        is_explicit: If True, this person's ID is explicitly defined in `people.yaml`.  You probably want to use [`make_explicit()`][acl_anthology.people.person.Person.make_explicit] rather than change this attribute.
    """

    id: str = field(on_setattr=[setters.validate, _update_person_index])
    parent: Anthology = field(repr=False, eq=False)
    _names: list[tuple[Name, NameLink]] = field(
        factory=list, converter=_name_list_converter
    )
    item_ids: list[AnthologyIDTuple] = field(
        factory=list, repr=lambda x: f"<list of {len(x)} AnthologyIDTuple objects>"
    )
    orcid: Optional[str] = field(
        default=None,
        on_setattr=[setters.validate, _update_person_index],
    )  # validator defined below
    comment: Optional[str] = field(default=None)
    degree: Optional[str] = field(default=None)
    similar_ids: list[str] = field(factory=list)
    disable_name_matching: Optional[bool] = field(default=False, converter=bool)
    is_explicit: Optional[bool] = field(default=False, converter=bool)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Person):
            return False
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    @orcid.validator
    def _check_orcid(self, _: Any, value: Optional[str]) -> None:
        if value is not None and not is_valid_orcid(value):
            raise ValueError(f"ORCID is not valid (wrong format or checksum): {value}")

    @property
    def names(self) -> list[Name]:
        return [name for (name, _) in self._names]

    @names.setter
    def names(self, values: list[Name]) -> None:
        for name, _ in self._names:
            self.parent.people._remove_name(self.id, name)
        for name in values:
            self.parent.people._add_name(self.id, name)
        self._names = _name_list_converter(values)

    @property
    def canonical_name(self) -> Name:
        """
        Returns:
            The canonical name for this person.
        """
        try:
            # By convention, the first entry of `self.names` is treated as the
            # canonical entry
            return self._names[0][0]
        except IndexError:
            raise ValueError(f"No names defined for person '{self.id}'")

    @canonical_name.setter
    def canonical_name(self, name: Name) -> None:
        self._set_canonical_name(name)

    def _set_canonical_name(self, name: Name, inferred: bool = False) -> None:
        """Set the canonical name for this person.

        Outside of the library, use Person.canonical_name = ...

        Parameters:
            name: Name that should be treated as canonical for this person.
            inferred: Marks the canonical name as inferred (used inside the name slug matching algorithm).
        """
        link_type = NameLink.INFERRED if inferred else NameLink.EXPLICIT
        if not self.has_name(name):
            self._names.insert(0, (name, link_type))
        else:
            self._names = [(name, link_type)] + [x for x in self._names if x[0] != name]

    def add_name(self, name: Name, inferred: bool = False) -> None:
        """Add a name for this person.

        Parameters:
            name: Name that can refer to this person.
            inferred: If True, will be marked as `NameLinkingType.INFERRED`, which will e.g. cause this name to not be written to `people.yaml`.  Used when building the [`PersonIndex`][acl_anthology.people.index.PersonIndex] from the XML data; you probably don't want to set this manually.  Defaults to False.
        """
        link_type = NameLink.INFERRED if inferred else NameLink.EXPLICIT
        if not self.has_name(name):
            self._names.append((name, link_type))
            self.parent.people._add_name(self.id, name)
        elif (name, link_type) not in self._names:
            # ensure that name is re-inserted at same position
            idx = self.names.index(name)
            del self._names[idx]
            self._names.insert(idx, (name, link_type))

    def remove_name(self, name: Name) -> None:
        """Remove an explicit name for this person.

        Warning:
            If the name is still used on a paper or volume with the ID of this person, this may result in an Exception during index building.  Names that were implicitly linked to this person cannot be removed this way, as the name would simply reappear on next load.

        Parameters:
            name: Name that should be removed from this person.

        Raises:
            ValueError: If this name was not explicitly linked to this person.
        """
        self._names.remove((name, NameLink.EXPLICIT))
        self.parent.people._remove_name(self.id, name)

    def has_name(self, name: Name) -> bool:
        """
        Parameters:
            name: Name to be checked.

        Returns:
            True if the given name can refer to this person.
        """
        return any(existing_name == name for (existing_name, _) in self._names)

    def change_id(self, new_id: str) -> None:
        """Change this person's ID.

        This updates `self.id`, but also ensures that all papers/items with the old ID are updated to the new one.

        Parameters:
            new_id: The new ID for this person, which must match [`RE_VERIFIED_PERSON_ID`][acl_anthology.utils.ids.RE_VERIFIED_PERSON_ID].

        Raises:
            AnthologyException: If `self.explicit` is False.
            AnthologyInvalidIDError: If the supplied ID is not well-formed, or if it already exists in the PersonIndex.
        """
        if new_id in self.parent.people:
            exc = AnthologyInvalidIDError(new_id, f"Person ID already exists: {new_id}")
            exc.add_note("Did you want to use merge_with_explicit() instead?")
            raise exc
        if not self.is_explicit:
            exc2 = AnthologyException("Can only update ID for explicit person")
            exc2.add_note("Did you want to use make_explicit() instead?")
            raise exc2
        if not is_verified_person_id(new_id):
            raise AnthologyInvalidIDError(
                new_id, f"Not a valid verified-person ID: {new_id}"
            )

        self._set_id_on_items(new_id)
        self.id = new_id  # triggers update in PersonIndex

    def make_explicit(self, new_id: str) -> None:
        """Turn this person that was implicitly created into an explicitly-represented one.

        This will result in this person having an explicit entry in `people.yaml` with all names that are currently associated with this person.  It will also add their new explicit ID to all papers and volumes currently associated with this person.

        Parameters:
            new_id: The new ID for this person, which must match [`RE_VERIFIED_PERSON_ID`][acl_anthology.utils.ids.RE_VERIFIED_PERSON_ID].

        Raises:
            AnthologyException: If `self.explicit` is already True.
            ValueError: If the supplied ID is not valid, or if it already exists in the PersonIndex.
        """
        if self.is_explicit:
            raise AnthologyException(f"Person '{self.id}' is already explicit")
        if not is_verified_person_id(new_id):
            raise AnthologyInvalidIDError(
                new_id, f"Not a valid verified-person ID: {new_id}"
            )

        self._set_id_on_items(new_id)
        self.is_explicit = True
        self.id = new_id  # triggers update in PersonIndex
        self._names = [(name, NameLink.EXPLICIT) for name, _ in self._names]

    def merge_with_explicit(self, person: Person) -> None:
        """Merge this person that was implicitly created with an explicitly-represented one.

        This will add the explicit person's ID to all papers and volumes currently associated with this inferred person.

        Parameters:
            person: An explicit person to merge this person's items into.

        Raises:
            AnthologyException: If `self.explicit` is True or `person.explicit` is False.
        """
        if self.is_explicit:
            raise AnthologyException("Can only merge non-explicit persons")
        if not person.is_explicit:
            raise AnthologyException(
                f"Can only merge with explicit persons; not '{person.id}'"
            )

        self._set_id_on_items(person.id)
        person.item_ids.extend(self.item_ids)
        self.item_ids = []

        for name in self.names:
            person.add_name(name, inferred=False)

    def _set_id_on_items(self, new_id: str) -> None:
        """Set `new_id` on all name specifications that currently resolve to this person.  Intended for internal use."""
        if self.is_explicit:

            def namespec_refers_to_self(namespec: NameSpecification) -> bool:
                return namespec.id == self.id

        else:

            def namespec_refers_to_self(namespec: NameSpecification) -> bool:
                return self.parent.resolve(namespec) is self

        for item in self.anthology_items():
            for namespec in item.namespecs:
                if namespec_refers_to_self(namespec):
                    namespec.id = new_id
                    item.collection.is_modified = True

    def anthology_items(self) -> Iterator[Paper | Volume]:
        """Returns an iterator over all Anthology items associated with this person, regardless of their type."""
        for anthology_id in self.item_ids:
            item = self.parent.get(anthology_id)
            if item is None:
                raise ValueError(
                    f"Person {self.id} lists associated item {build_id_from_tuple(anthology_id)}, which doesn't exist"
                )  # pragma: no cover
            # TODO: typing issue will be resolved later with CollectionItem refactoring
            yield item  # type: ignore

    def papers(self) -> Iterator[Paper]:
        """Returns an iterator over all papers associated with this person.

        Note:
            This will return papers where this person is an author, as well as frontmatter of volumes where they are an editor. It will _not_ include all other papers in volumes they have edited.
        """
        for anthology_id in self.item_ids:
            paper_id = anthology_id[-1]
            if paper_id is not None:
                paper = self.parent.get_paper(anthology_id)
                if paper is None:
                    raise ValueError(
                        f"Person {self.id} lists associated paper {build_id_from_tuple(anthology_id)}, which doesn't exist"
                    )  # pragma: no cover
                yield paper

    def volumes(self) -> Iterator[Volume]:
        """Returns an iterator over all volumes this person has edited."""
        for anthology_id in self.item_ids:
            paper_id = anthology_id[-1]
            if paper_id is None:
                volume = self.parent.get_volume(anthology_id)
                if volume is None:
                    raise ValueError(
                        f"Person {self.id} lists associated volume {build_id_from_tuple(anthology_id)}, which doesn't exist"
                    )  # pragma: no cover
                yield volume



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for XML serialization."""

import itertools as it
from difflib import SequenceMatcher
from lxml import etree
from typing import Callable, Iterable, Optional
from xml.sax.saxutils import escape as xml_escape

TAGS_WITH_MARKUP = {
    "b",
    "i",
    "fixed-case",
    "title",
    "abstract",
    "booktitle",
    "shortbooktitle",
}
"""XML tags which contain MarkupText."""


TAGS_WITH_UNORDERED_CHILDREN = {
    "talk",
    "paper",
    "meta",
    "frontmatter",
    "event",
    "colocated",
    "author",
    "editor",
    "speaker",
    "variant",
}
"""XML tags that may contain child elements that can logically appear in arbitrary order."""


TAGS_WITH_ORDER_SEMANTICS = {
    "author",
    "editor",
    "speaker",
    "erratum",
    "revision",
    "talk",
    "venue",
    # These maybe shouldn't matter, but currently do:
    "attachment",
    "award",
    "video",
}
"""XML tags that may appear multiple times per parent tag, and whose relative order matters even if their parent tag belongs to `TAGS_WITH_UNORDERED_CHILDREN`."""


TAGS_WITHOUT_LINEBREAKS = {
    "author",
    "editor",
    "speaker",
    "title",
    "booktitle",
    "variant",
}
"""XML tags that should always be serialized without line breaks."""


def _filter_children(elems: Iterable[etree._Element]) -> list[etree._Element]:
    """Filter child elements that contribute no information.

    This includes superfluous empty XML elements, such as <first/> tags, as well as XML comments.
    """
    taglist = (
        "abstract",
        "address",
        "affiliation",
        "colocated",
        "dates",
        "first",
        "pages",
        "publisher",
    )
    return [
        x
        for x in elems
        if not (x.tag in taglist and not x.text) and not isinstance(x, etree._Comment)
    ]


def _sort_children(x: etree._Element) -> tuple[str, str]:
    """Turn an XML element into a key for sorting purposes."""
    if x.tag in TAGS_WITH_ORDER_SEMANTICS:
        # We return the same sorting key for all tags whose relative order
        # should not be changed. This guarantees that their original order will
        # not be changed due to Python's sort stability:
        # <https://docs.python.org/3/howto/sorting.html#sort-stability-and-complex-sorts>
        return (x.tag, "")
    return (x.tag, etree.tostring(x, encoding="unicode"))


def assert_equals(elem: etree._Element, other: etree._Element) -> None:
    """Assert that two Anthology XML elements are logically equivalent.

    Arguments:
        elem: The first element to compare.
        other: The second element to compare.

    Raises:
        AssertionError: If the two elements are not logically equivalent.
    """
    assert elem.tag == other.tag
    assert elem.attrib == other.attrib
    assert elem.text == other.text or (not elem.text and not other.text)
    if elem.tag in TAGS_WITH_MARKUP:
        # Should render identically, except maybe for trailing whitespace
        assert (
            etree.tostring(elem, encoding="unicode").rstrip()
            == etree.tostring(other, encoding="unicode").rstrip()
        )
    else:
        elem_children = _filter_children(elem)
        other_children = _filter_children(other)
        if elem_children and elem.tag in TAGS_WITH_UNORDERED_CHILDREN:
            elem_children = sorted(elem_children, key=_sort_children)
            other_children = sorted(other_children, key=_sort_children)
        assert [child.tag for child in elem_children] == [
            child.tag for child in other_children
        ]
        for elem_child, other_child in zip(elem_children, other_children):
            assert_equals(elem_child, other_child)


def append_text(elem: etree._Element, text: str) -> None:
    """Append text to an XML element.

    If the XML element has children, the text will be appended to the tail of the last child; otherwise, it will be appended to its text attribute.

    Arguments:
        elem: The XML element.
        text: The text string to append to the XML element.

    Returns:
        None; the XML element is modified in-place.
    """
    if len(elem):
        # already has children — append text to tail
        if elem[-1].tail is not None:
            elem[-1].tail = "".join((elem[-1].tail, text))
        else:
            elem[-1].tail = text
    elif elem.text is not None:
        elem.text = "".join((elem.text, text))
    else:
        elem.text = text


def clean_whitespace(
    text: Optional[str], func: Optional[Callable[[str], str]] = None
) -> Optional[str]:
    if text is None:
        return text
    while "  " in text:
        text = text.replace("  ", " ")
    if func is not None:
        text = func(text)
    return text


def indent(elem: etree._Element, level: int = 0, internal: bool = False) -> None:
    """Enforce canonical indentation.

    "Canonical indentation" is two spaces, with each tag on a new line,
    except that 'author', 'editor', 'title', and 'booktitle' tags are
    placed on a single line.

    Arguments:
        elem: The XML element to apply canonical indentation to.
        level: Indentation level; used for recursive calls of this function.
        internal: If True, assume we are within a single-line element.

    Note:
        Adapted from [https://stackoverflow.com/a/33956544](https://stackoverflow.com/a/33956544).
    """
    if isinstance(elem, etree._Comment):
        return

    # tags that have no internal linebreaks (including children)
    oneline = elem.tag in TAGS_WITHOUT_LINEBREAKS

    elem.text = clean_whitespace(elem.text, lambda x: x.lstrip())
    is_markup = elem.tag in TAGS_WITH_MARKUP

    if len(elem):  # children
        # Set indent of first child for tags with no text
        if not oneline and not is_markup and (not elem.text or not elem.text.strip()):
            elem.text = "\n" + (level + 1) * "  "

        if not elem.tail or not elem.tail.strip():
            if level:
                elem.tail = "\n" + level * "  "
            else:
                elem.tail = "\n"

        if is_markup:
            child = elem[-1]
        else:
            # recurse
            for child in elem:
                indent(
                    child,
                    level + 1,
                    internal=(internal or oneline),
                )

        # Clean up the last child
        if oneline:
            child.tail = clean_whitespace(child.tail, lambda x: x.rstrip())
        elif (
            not is_markup and not internal and (not child.tail or not child.tail.strip())
        ):
            child.tail = "\n" + level * "  "
    else:
        elem.text = clean_whitespace(elem.text, lambda x: x.strip())

        if internal:
            elem.tail = clean_whitespace(elem.tail)
        elif not elem.tail or not elem.tail.strip():
            elem.tail = "\n" + level * "  "


def ensure_minimal_diff(elem: etree._Element, reference: etree._Element) -> None:
    """Change a node to minimize the diff compared to a reference node, without changing logical equivalence.

    This will change the order of nodes and attributes to match the order in the reference whenever this makes no functional difference.  Elements that are logically equivalent to those in the reference will be copied exactly.

    Arguments:
        elem: The XML element whose children should be matches.
        reference: The XML element that serves as a reference.

    Raises:
        ValueError: If `elem` and `reference` do not have identical tags.
    """
    if elem.tag != reference.tag:
        raise ValueError(
            f"ensure_minimal_diff received two elements with different tags ({elem.tag} != {reference.tag})"
        )

    # If the entire elements are logically equivalent, we just clone the reference
    try:
        assert_equals(elem, reference)
        # NOTE: not using elem.getparent().replace() here since that breaks
        # recursion (it also removes the reference from its parent)
        elem.clear()
        for key, value in reference.items():
            elem.set(key, value)
        elem.text = reference.text
        elem.tail = reference.tail
        elem.extend(reference)
        return
    except AssertionError:
        pass  # Need to do more work

    # Sort attributes to match reference
    if len(elem.attrib) > 1 and len(reference.attrib) > 1:
        # Follow key order in reference, with keys new in elem coming last
        attrib_order = list(it.chain(reference.keys(), elem.keys()))
        # Since elem.attrib is an etree._Attrib, not a plain dictionary, we
        # clear attribs and reassign in the desired order
        attribs = sorted(elem.items(), key=lambda x: attrib_order.index(x[0]))
        elem.attrib.clear()
        for key, value in attribs:
            elem.set(key, value)

    # Sort child elements to match order in reference, if element allows reordering
    if elem.tag in TAGS_WITH_UNORDERED_CHILDREN:
        # Follow tag order in reference, with keys new in elem coming last
        ref_order = [x.tag for x in it.chain(reference, elem)]
        # Sort the child elements by the given reference order, reassigning
        # them to the parent element.  Since the sort key is only the tag,
        # elements will the same tag will always be grouped together, and
        # Python's sort stability guarantees that this does not change the
        # relative order of elements with the same tag.
        elem[:] = sorted(elem, key=lambda x: ref_order.index(x.tag))

    # From here on, children of elem have either been reordered to match
    # reference, or the order carries meaning.  Since elem may have added or
    # deleted child elements compared to reference, we now use difflib to find
    # "corresponding" child elements to recurse into.  Two elements are
    # considered as "corresponding" to each other if they have the same tag and
    # attributes.
    def make_match_keys(elems: Iterable[etree._Element]) -> list[str]:
        """Key used for element matching."""
        return [f"{elem.tag}{sorted(elem.items())}" for elem in elems]

    matcher = SequenceMatcher(
        a=make_match_keys(elem),
        b=make_match_keys(reference),
    )
    for a, b, size in matcher.get_matching_blocks():
        for idx_elem, idx_ref in zip(range(a, a + size), range(b, b + size)):
            ensure_minimal_diff(elem[idx_elem], reference[idx_ref])


def stringify_children(node: etree._Element) -> str:
    """
    Arguments:
        node: An XML element.

    Returns:
        The full content of the input node, including tags.

    Used for nodes that can have mixed text and HTML elements (like `<b>` and `<i>`).
    """
    return "".join(
        chunk
        for chunk in it.chain(
            (xml_escape_or_none(node.text),),
            it.chain(
                *(
                    (
                        etree.tostring(child, with_tail=False, encoding=str),
                        xml_escape_or_none(child.tail),
                    )
                    for child in node
                )
            ),
            (xml_escape_or_none(node.tail),),
        )
        if chunk
    ).strip()


def xml_escape_or_none(t: Optional[str]) -> Optional[str]:
    """Like [xml.sax.saxutils.escape][], but accepts [None][]."""
    return None if t is None else xml_escape(t)


def xsd_boolean(value: str) -> bool:
    """Converts an xsd:boolean value to a bool."""
    if value in ("0", "false"):
        return False
    elif value in ("1", "true"):
        return True
    raise ValueError(f"Not a valid xsd:boolean value: '{value}'")



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions implementing the automatic cloning of the Anthology repository."""

from __future__ import annotations

from git.repo import Repo
from git import RemoteProgress
from pathlib import Path
from rich.progress import Progress
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import StrPath

from .logging import get_logger

log = get_logger()


def clone_or_pull_from_repo(
    repo_url: str, local_path: StrPath, verbose: Optional[bool]
) -> None:
    """Clones a Git repository, or pulls from remote if it already exists.

    Arguments:
        repo_url: The URL of a Git repo.
        local_path: The local path containing the repo.  If `{local_path}/.git` exists, we assume it already contains the repo and will attempt to pull from 'origin'; otherwise, we will attempt to clone the repo into `local_path`.
        verbose: If True, will show a progress display.
    """
    path = Path(local_path)
    progress = RichRemoteProgress() if verbose else None
    log.debug(f"Using local repository folder: {path}")
    if (path / ".git").exists():
        repo = Repo(path)
        if repo.remote().url != repo_url:
            log.error(
                (
                    "Repository folder exists, but doesn't match the given URL:\n",
                    f"   {repo.remote().url} != {repo_url}",
                )
            )
        log.info(f"Fetching updates from: {repo_url}")
        repo.remote().pull(force=True, progress=progress)
    else:
        log.info(f"Cloning repository: {repo_url}")
        repo = Repo.clone_from(
            repo_url, path, progress=progress, single_branch=True, depth=1  # type: ignore[arg-type]
        )
        # ^-- It seems that Repo.clone_from() has an incorrect type signature
        # for its progress argument...
    if progress is not None:
        progress.progress.update(progress.task, completed=210.0)
        progress.progress.stop()


# This is a bit hacky, mainly to provide a "smoother" user experience, but
# could probably be simplified.
class RichRemoteProgress(RemoteProgress):
    def __init__(self) -> None:
        super().__init__()
        self.progress = Progress()
        self.task = self.progress.add_task(
            "Fetching Anthology data...", start=False, visible=False, total=210.0
        )

    def update(
        self,
        op_code: int,
        cur_count: str | float,
        max_count: Optional[str | float] = None,
        message: str = "",
    ) -> None:
        operation = op_code & self.OP_MASK

        # Satisfying the type checker, even though this _should_ always be float ...
        if isinstance(cur_count, str):
            try:
                cur_count = float(cur_count)
            except ValueError:
                cur_count = 0.0
        if isinstance(max_count, str):
            try:
                max_count = float(max_count)
            except ValueError:
                max_count = 100.0
        elif max_count is None:
            max_count = 100.0

        if operation == self.COUNTING:
            if op_code & self.BEGIN:
                self.progress.start()
                self.progress.start_task(self.task)
                self.progress.update(self.task, visible=True)
            if op_code & self.END:
                value = 9.0
                self.progress.update(self.task, completed=value)
        elif operation == self.COMPRESSING:
            value = 9.0 + 100.0 * (cur_count / max_count)
            self.progress.update(self.task, completed=value)
        elif operation == self.RECEIVING:
            value = 109.0 + 100.0 * (cur_count / max_count)
            self.progress.update(self.task, completed=value)
        elif (op_code & self.END) and (operation == self.RESOLVING):
            self.progress.update(self.task, completed=210.0)
            self.progress.stop()



# Copyright 2019-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes and functions for text markup manipulation."""

from __future__ import annotations

from attrs import define, field, validators as v
from collections import defaultdict
from copy import deepcopy
from functools import total_ordering
from lxml import etree
from typing import Iterator, Optional, SupportsIndex
from xml.sax.saxutils import escape as xml_escape

from ..utils import (
    clean_unicode,
    latex_encode,
    latex_convert_quotes,
    parse_latex_to_xml,
    remove_extra_whitespace,
    stringify_children,
)
from .texmath import TexMath

MARKUP_LATEX_CMDS = defaultdict(
    lambda: "{text}",
    {
        "fixed-case": "{{{text}}}",
        "b": "\\textbf{{{text}}}",
        "i": "\\textit{{{text}}}",
        "tex-math": "${text}$",
        "url": "\\url{{{text}}}",
    },
)


def markup_to_latex(element: etree._Element) -> str:
    tag = str(element.tag)
    if tag in ("tex-math", "url"):
        # These tags cannot have child elements
        return MARKUP_LATEX_CMDS[tag].format(text=element.text)

    text = latex_encode(element.text)
    for nested_element in element:
        text += markup_to_latex(nested_element)
        if nested_element.tail:
            text += latex_encode(nested_element.tail)

    text = MARKUP_LATEX_CMDS[tag].format(text=text)
    return text


@define(repr=False)
@total_ordering
class MarkupText:
    """Text with optional markup.

    Warning:
        This class **should not be instantiated directly.**  Use its class method constructors instead.

    Example:
        ```python
        title = MarkupText.from_("A Structured Review of the Validity of BLEU")
        title = MarkupText.from_("TTCS$^{\\mathcal{E}}$: a Vectorial Resource for Computing Conceptual Similarity")
        ```

    Note:
        This class implements a limited subset of string methods to make common operations more convenient, for example:

        ```python
        title = MarkupText.from_string("A Structured Review of the Validity of BLEU")
        title == "A Structured Review of the Validity of BLEU"  # True
        "BLEU" in title  # True
        title.startswith("A ")  # True
        ```

        These operate on the [stringified XML representation][acl_anthology.text.markuptext.MarkupText.as_xml] of the class.
    """

    # IMPLEMENTATION NOTE: Deepcopy-ing (or newly instantiating) etree._Element
    # is very expensive, as shown by profiling. Therefore, markup elements
    # which don't actually contain any markup are simply stored as
    # strings. This makes the implementation slightly more verbose (we need to
    # check everywhere whether we're dealing with etree._Element or str), but
    # much faster. ---For further optimization, we could explore if there's an
    # alternative that doesn't require deepcopy-ing XML elements at all.
    _content: etree._Element | str = field(validator=v.instance_of((etree._Element, str)))

    # For caching
    _html: Optional[str] = field(init=False, default=None)
    _latex: Optional[str] = field(init=False, default=None)
    _text: Optional[str] = field(init=False, default=None)
    _xml: Optional[str] = field(init=False, default=None)

    def __contains__(self, key: object) -> bool:
        if isinstance(key, str):
            return (key in self.as_text()) or (key in self.as_xml())
        return False  # pragma: no cover

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.as_xml() == other
        elif isinstance(other, MarkupText):
            return self.as_xml() == other.as_xml()
        return False  # pragma: no cover

    def __lt__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.as_xml() < other
        elif isinstance(other, MarkupText):
            return self.as_xml() < other.as_xml()
        return False  # pragma: no cover

    def __len__(self) -> int:
        return len(self.as_text())

    def __str__(self) -> str:
        return self.as_text()

    def __repr__(self) -> str:
        return f"<MarkupText {self.as_html()!r}>"

    def __rich_repr__(self) -> Iterator[str]:
        yield self.as_xml()

    @property
    def contains_markup(self) -> bool:
        """True if this text contains markup; False if it is a plain string."""
        return isinstance(self._content, etree._Element)

    def as_text(self) -> str:
        """
        Returns:
            The plain text with any markup stripped. The only transformation that will be performed is replacing TeX-math expressions with their corresponding Unicode representation, if possible.
        """
        if isinstance(self._content, str):
            return remove_extra_whitespace(self._content)
        if self._text is not None:
            return self._text
        element = deepcopy(self._content)
        for sub in element.iterfind(".//tex-math"):
            sub.text = TexMath.to_unicode(sub)
        text = etree.tostring(element, encoding="unicode", method="text")
        self._text = remove_extra_whitespace(text)
        return self._text

    def as_html(self, allow_url: bool = True) -> str:
        """
        Returns:
            Text with markup transformed into HTML.

        Arguments:
            allow_url: Defaults to True. If False, URLs are **not** wrapped in
                `<a href="...">` tags, but in simply `<span>` tags.
        """
        if isinstance(self._content, str):
            return xml_escape(remove_extra_whitespace(self._content))
        if self._html is not None:
            return self._html
        element = deepcopy(self._content)
        for sub in element.iter():
            if sub.tag == "url":
                if allow_url:
                    sub.tag = "a"
                    sub.set("href", str(sub.text))
                else:
                    sub.tag = "span"
                sub.set("class", "acl-markup-url")
            elif sub.tag == "fixed-case":
                sub.tag = "span"
                sub.set("class", "acl-fixed-case")
            elif sub.tag == "tex-math":
                parsed_elem = TexMath.to_html(sub)
                parsed_elem.tail = sub.tail
                sub.getparent().replace(sub, parsed_elem)  # type: ignore
            elif len(sub) == 0 and sub.text is None:
                sub.text = ""
        self._html = remove_extra_whitespace(stringify_children(element))
        return self._html

    def as_latex(self) -> str:
        """
        Returns:
            Text with markup transformed into LaTeX commands.
        """
        if self._latex is not None:
            return self._latex
        if isinstance(self._content, str):
            latex = latex_encode(self._content)
        else:
            latex = markup_to_latex(self._content)
        self._latex = remove_extra_whitespace(latex_convert_quotes(latex))
        return self._latex

    def as_xml(self) -> str:
        """
        Returns:
            Text with markup in the original XML format.
        """
        if isinstance(self._content, str):
            return xml_escape(self._content)
        if self._xml is None:
            self._xml = stringify_children(self._content)
        return self._xml

    @classmethod
    def from_string(cls, text: str, clean: bool = True) -> MarkupText:
        """
        Arguments:
            text: A simple text string without any markup.
            clean: If True, applies the Anthology's [Unicode normalization][acl_anthology.utils.text.clean_unicode].

        Returns:
            Instantiated MarkupText object corresponding to the string.
        """
        if clean:
            return cls(clean_unicode(text))
        return cls(text)

    @classmethod
    def from_latex(cls, text: str, clean: bool = True) -> MarkupText:
        """
        Arguments:
            text: A text string potentially containing LaTeX markup.
            clean: If True, applies the Anthology's [Unicode normalization][acl_anthology.utils.text.clean_unicode].

        Returns:
            Instantiated MarkupText object corresponding to the string.
        """
        if clean:
            text = clean_unicode(text)
        element = parse_latex_to_xml(text, use_heuristics=False)
        return cls.from_xml(element)

    @classmethod
    def from_latex_maybe(cls, text: str, clean: bool = True) -> MarkupText:
        """
        Like `from_latex()`, but can be used if it is unclear if the string is plain text or LaTeX.  Will prevent percentage signs being interpreted as LaTeX comments, and apply a heuristic to decide if a tilde is literal or a non-breaking space.

        Arguments:
            text: A text string potentially in plain text or LaTeX format.
            clean: If True, applies the Anthology's [Unicode normalization][acl_anthology.utils.text.clean_unicode].

        Returns:
            Instantiated MarkupText object corresponding to the string.
        """
        if clean:
            text = clean_unicode(text)
        element = parse_latex_to_xml(text, use_heuristics=True)
        return cls.from_xml(element)

    @classmethod
    def from_xml(cls, element: etree._Element) -> MarkupText:
        """
        Arguments:
            element: An XML element containing valid MarkupText according to the schema.

        Returns:
            Instantiated MarkupText object corresponding to the element.
        """
        if len(element):
            return cls(deepcopy(element))
        return cls(str(element.text) if element.text is not None else "")

    @classmethod
    def from_(cls, content: etree._Element | str) -> MarkupText:
        """Instantiate MarkupText from an XML element or a string, heuristically parsing any supported markup.

        - If called with an XML element, assumes it uses the Anthology's markup format and calls [`from_xml()`][acl_anthology.text.markuptext.MarkupText.from_xml].
        - If called with a string, assumes the string might contain markup and will try to intelligently parse it.  At the moment, only LaTeX markup is supported, which means that the effect of this is identical to calling [`from_latex_maybe()`][acl_anthology.text.markuptext.MarkupText.from_latex_maybe], but this may change if we support different types of markup in the future.

        Note:
            If you want more fine-grained control over markup detection, call one of the more specific builder functions instead.

        Arguments:
            content: A string potentially containing markup, or an XML element containing valid MarkupText according to the schema.

        Returns:
            Instantiated MarkupText object corresponding to the content.
        """
        if isinstance(content, etree._Element):
            return cls.from_xml(content)
        elif isinstance(content, str):
            return cls.from_latex_maybe(content)
        else:  # pragma: no cover
            raise TypeError(f"Cannot instantiate MarkupText from {type(content)}")

    def to_xml(self, tag: str = "span") -> etree._Element:
        """
        Arguments:
            tag: Name of outer tag in which the text should be wrapped.

        Returns:
            A serialization of this MarkupText in Anthology XML format.
        """
        if isinstance(self._content, str):
            element = etree.Element(tag)
            if self._content:
                element.text = self._content
        else:
            element = deepcopy(self._content)
            element.tag = tag
        return element

    ### STRING METHODS

    def endswith(
        self,
        suffix: str | tuple[str, ...],
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> bool:
        """Return True if the string ends with the specified suffix, False otherwise.

        Equivalent to `self.as_xml().endswith(...)`.
        """
        return self.as_xml().endswith(suffix, start, end)

    def startswith(
        self,
        prefix: str | tuple[str, ...],
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> bool:
        """Return True if the string starts with the specified prefix, False otherwise.

        Equivalent to `self.as_xml().startswith(...)`.
        """
        return self.as_xml().startswith(prefix, start, end)


def to_markuptext(value: object) -> MarkupText:
    if isinstance(value, MarkupText):
        return value
    # MarkupText.from_ already raises TypeError, so let's not duplicate that here
    return MarkupText.from_(value)  # type: ignore[arg-type]



# Copyright 2023-2026 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from attrs import define, field
from collections import Counter, defaultdict
import itertools as it
from pathlib import Path
from rich.progress import track
from scipy.cluster.hierarchy import DisjointSet  # type: ignore
import sys
from typing import cast, Any, Iterable, Optional, TYPE_CHECKING
import warnings
import yaml

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:  # pragma: no cover
    from yaml import Loader, Dumper  # type: ignore

from ..config import primary_console
from ..containers import SlottedDict
from ..exceptions import (
    AnthologyException,
    AnthologyInvalidIDError,
    NameSpecResolutionError,
    NameSpecResolutionWarning,
    PersonDefinitionError,
)
from ..utils.ids import AnthologyIDTuple, is_verified_person_id
from ..utils.logging import get_logger
from . import Person, Name, NameLink, NameSpecification
from .name import _YAMLName

if TYPE_CHECKING:
    from _typeshed import StrPath
    from ..anthology import Anthology
    from ..collections import Paper, Volume, Collection

log = get_logger()
PEOPLE_INDEX_FILE = "yaml/people.yaml"


UNVERIFIED_PID_FORMAT = "{pid}/unverified"
"""Format string for unverified person IDs."""
# Note: Changing this will require changes in Hugo templates etc. as well!


@define
class PersonIndex(SlottedDict[Person]):
    """Index object through which all persons (authors/editors) can be accessed.

    Provides dictionary-like functionality mapping person IDs to [Person][acl_anthology.people.person.Person] objects.

    Info:
        All information about persons is currently derived from [name specifications][acl_anthology.people.name.NameSpecification] on volumes and papers, and not stored explicitly. This means:

        1. Loading this index requires parsing the entire Anthology data.
        2. Nothing in this index should be modified to make changes to Anthology data; change the information on papers instead.

        See the [guide on accessing author/editor information](../guide/accessing-authors.md) for more information.

    Attributes:
        parent: The parent Anthology instance to which this index belongs.
        path: The path to `people.yaml`.
        by_orcid: A mapping of ORCIDs (as strings) to person IDs.
        by_name: A mapping of [Name][acl_anthology.people.name.Name] instances to lists of person IDs.
        slugs_to_verified_ids: A mapping of strings (representing slugified names) to lists of person IDs.
        similar: A [disjoint-set structure][scipy.cluster.hierarchy.DisjointSet] of persons with similar names.
        is_data_loaded: A flag indicating whether the index has been constructed.
    """

    parent: Anthology = field(repr=False, eq=False)
    path: Path = field(init=False)
    _by_orcid: dict[str, str] = field(init=False, repr=False, default={})
    _by_name: dict[Name, list[str]] = field(
        init=False, repr=False, factory=lambda: defaultdict(list)
    )
    _slugs_to_verified_ids: dict[str, set[str]] = field(
        init=False, repr=False, factory=lambda: defaultdict(list)
    )
    _similar: DisjointSet = field(init=False, repr=False, factory=DisjointSet)
    is_data_loaded: bool = field(init=False, repr=True, default=False)

    @path.default
    def _path(self) -> Path:
        return self.parent.datadir / Path(PEOPLE_INDEX_FILE)

    @property
    def by_orcid(self) -> dict[str, str]:
        if not self.is_data_loaded:
            self.load()
        return self._by_orcid

    @property
    def by_name(self) -> dict[Name, list[str]]:
        if not self.is_data_loaded:
            self.load()
        return self._by_name

    @property
    def similar(self) -> DisjointSet:
        if not self.is_data_loaded:
            self.load()
        return self._similar

    @property
    def slugs_to_verified_ids(self) -> dict[str, set[str]]:
        if not self.is_data_loaded:
            self.load()
        return self._slugs_to_verified_ids

    def get_by_name(self, name: Name) -> list[Person]:
        """Access persons by their name.

        Parameters:
            name: A personal name.

        Returns:
            A list of all persons with that name; can be empty.
        """
        if not self.is_data_loaded:
            self.load()
        return [self.data[pid] for pid in self._by_name[name]]

    def get_by_namespec(self, name_spec: NameSpecification) -> Person:
        """Access persons by their name specification.

        See [resolve_namespec()][acl_anthology.people.index.PersonIndex.resolve_namespec] for exceptions that can be raised by this function.

        Parameters:
            name_spec: A name specification.

        Returns:
            The person associated with this name specification.
        """
        if not self.is_data_loaded:
            self.load()
        return self.resolve_namespec(name_spec)

    def get_by_orcid(self, orcid: str) -> Person | None:
        """Access persons by their ORCID.

        Parameters:
            orcid: A string representing an ORCID.

        Returns:
            The person with that ORCID, if it exists, otherwise None.
        """
        if not self.is_data_loaded:
            self.load()
        if orcid in self._by_orcid:
            return self.data[self._by_orcid[orcid]]
        return None

    def find_coauthors(
        self, person: str | Person, include_volumes: bool = True
    ) -> list[Person]:
        """Find all persons who co-authored or co-edited items with the given person.

        Parameters:
            person: A person ID _or_ Person instance.
            include_volumes: If set to False, will not consider co-editorship on volumes, unless they have frontmatter.

        Returns:
            A list of all persons who are co-authors; can be empty.
        """
        coauthors = self.find_coauthors_counter(person, include_volumes=include_volumes)
        return [self.data[pid] for pid in coauthors]

    def find_coauthors_counter(
        self, person: str | Person, include_volumes: bool = True
    ) -> Counter[str]:
        """Find the count of co-authored or co-edited items per person.

        Parameters:
            person: A person ID _or_ Person instance.
            include_volumes: If set to False, will not consider co-editorship on volumes, unless they have frontmatter.

        Returns:
            A Counter mapping **IDs** of other persons Y to the number of papers this person has co-authored with Y.
        """
        if not self.is_data_loaded:
            self.load()
        if isinstance(person, str):
            person = self.data[person]
        coauthors: Counter[str] = Counter()
        for item_id in person.item_ids:
            item = cast("Volume | Paper", self.parent.get(item_id))
            if (
                not include_volumes
                and item.full_id_tuple[-1] is None  # item is a Volume
                and not cast("Volume", item).has_frontmatter
            ):
                continue
            coauthors.update(self.resolve_namespec(ns).id for ns in item.editors)
            if hasattr(item, "authors"):
                coauthors.update(self.resolve_namespec(ns).id for ns in item.authors)
        del coauthors[person.id]
        return coauthors

    def load(self) -> None:
        """Loads or builds the index."""
        # This function exists so we can later add the option to read the index
        # from a cache if it doesn't need re-building.
        if self.is_data_loaded:  # pragma: no cover
            return
        self.build(show_progress=self.parent.verbose)

    def reset(self) -> None:
        """Resets the index."""
        self.data = {}
        self._by_orcid = {}
        self._by_name = defaultdict(list)
        self._slugs_to_verified_ids = defaultdict(set)
        self._similar = DisjointSet()
        self.is_data_loaded = False

    def build(self, show_progress: bool = False) -> None:
        """Load the entire Anthology data and build an index of persons.

        Important:
            Exceptions raised during the index creation are sent to the logger, and only a generic exception is raised at the end.
        """
        self.reset()
        self._load_people_index()
        # Go through every single volume/paper and add authors/editors
        if not show_progress:
            iterator: Iterable[Collection] = self.parent.collections.values()
        else:
            iterator = track(
                self.parent.collections.values(),
                total=len(self.parent.collections),
                description="Building person index...",
                console=primary_console,
            )
        raised_exception = False
        for collection in iterator:
            for volume in collection.volumes():
                context: Paper | Volume = volume
                try:
                    self._add_to_index(
                        volume.editors, volume.full_id_tuple, during_build=True
                    )
                    for paper in volume.papers():
                        context = paper
                        name_specs = (
                            # Associate explicitly given authors/editors with the paper
                            it.chain(paper.authors, paper.editors)
                            # For frontmatter, also associate the volume editors with it
                            if not paper.is_frontmatter
                            else paper.get_editors()
                        )
                        self._add_to_index(
                            name_specs, paper.full_id_tuple, during_build=True
                        )
                except Exception as exc:  # pragma: no cover
                    note = f"Raised in {context.__class__.__name__} {context.full_id}"
                    # If this is merged into a single if-statement (with "or"),
                    # the type checker complains ¯\_(ツ)_/¯
                    if isinstance(exc, AnthologyException):
                        exc.add_note(note)
                    elif sys.version_info >= (3, 11):
                        exc.add_note(note)
                    log.exception(exc)
                    raised_exception = True
        if raised_exception:
            raise Exception(
                "An exception was raised while building PersonIndex; check the logger for details."
            )  # pragma: no cover
        self.is_data_loaded = True

    def _load_people_index(self) -> None:
        """Load and parse the `people.yaml` file.

        Raises:
            AnthologyInvalidIDError: If `people.yaml` contains a malformed person ID; or if a person is listed without any names.
        """
        with open(self.path, "r", encoding="utf-8") as f:
            data = yaml.load(f, Loader=Loader)

        for pid, entry in data.items():
            if not is_verified_person_id(pid):
                raise AnthologyInvalidIDError(
                    pid, f"Invalid person ID in people.yaml: {pid}"
                )  # pragma: no cover
            self.add_person(
                Person(
                    id=pid,
                    parent=self.parent,
                    names=[Name.from_dict(n) for n in entry.pop("names")],
                    orcid=entry.pop("orcid", None),
                    comment=entry.pop("comment", None),
                    degree=entry.pop("degree", None),
                    similar_ids=entry.pop("similar", []),
                    disable_name_matching=entry.pop("disable_name_matching", False),
                    is_explicit=True,
                )
            )

            # Check for unprocessed keys to catch errors
            if entry:
                log.warning(
                    f"people.yaml: entry '{pid}' has unknown keys: {entry.keys()}"
                )  # pragma: no cover

    def add_person(self, person: Person) -> None:
        """Add a new person to the index.

        Parameters:
            person: The person to add, which should not exist in the index yet.

        Raises:
            AnthologyInvalidIDError: If a person with the same ID or ORCID already exists in the index.
        """
        if (pid := person.id) in self.data:
            raise AnthologyInvalidIDError(
                pid, f"A Person with ID '{pid}' already exists in the index"
            )
        self.data[pid] = person
        self._similar.add(pid)
        if person.orcid is not None:
            if person.orcid in self._by_orcid:
                raise ValueError(
                    f"ORCID '{person.orcid}' already assigned to person '{self._by_orcid[person.orcid]}'"
                )
            self._by_orcid[person.orcid] = pid
        for name in person.names:
            self._add_name(pid, name, during_build=True)
        for similar_id in person.similar_ids:
            self._similar.add(similar_id)  # might not have been added yet
            self._similar.merge(pid, similar_id)

    def create(
        self,
        id: str,
        names: list[Name],
        **kwargs: Any,
    ) -> Person:
        """Create a new explicit person and add it to the index.

        Parameters:
            id: The ID of the new person.
            names: A list of names for the new person; must contain at least one.
            **kwargs: Any valid list or optional attribute of [Person][acl_anthology.people.person.Person].

        Returns:
            The created [Person][acl_anthology.people.person.Person] object.

        Raises:
            AnthologyInvalidIDError: If a person with the given ID already exists, or the ID is not a well-formed verified-person ID.
            ValueError: If the list of names is empty.
        """
        if not self.is_data_loaded:
            self.load()
        if id in self.data:
            raise AnthologyInvalidIDError(
                id, f"A Person with ID '{id}' already exists in the index"
            )
        if not is_verified_person_id(id):
            raise AnthologyInvalidIDError(id, f"Not a valid verified-person ID: {id}")
        if not names:
            raise ValueError("List of names cannot be empty")

        kwargs["parent"] = self.parent
        kwargs["is_explicit"] = True

        person = Person(id=id, names=names, **kwargs)
        self.add_person(person)
        return person

    def _update_id(self, old_id: str, new_id: str) -> None:
        """Update a person ID in the index.

        Will change all indices to remove the old ID and replace it with the new one.  Will be called automatically from Person; do not call manually.

        Parameters:
            old_id: A person ID that already exists in the index.
            new_id: The new person ID it should be changed to, which mustn't exist in the index.

        Raises:
            KeyError: If the new ID already exists.
        """
        if not self.is_data_loaded:
            return
        if new_id in self.data:
            raise KeyError(
                f"Tried to add ID '{new_id}' to PersonIndex which already exists"
            )
        person = self.data.pop(old_id)
        self.data[new_id] = person
        # Note: cannot remove from DisjointSet
        self._similar.add(new_id)
        self._similar.merge(old_id, new_id)
        if person.orcid is not None:
            self._by_orcid[person.orcid] = new_id
        for name in person.names:
            self._remove_name(old_id, name)
            self._add_name(new_id, name)

    def _update_orcid(self, pid: str, old: Optional[str], new: Optional[str]) -> None:
        """Update a person's ORCID in the index.

        Will be called automatically from Person; do not call manually.
        """
        if not self.is_data_loaded:
            return
        if old is not None and old in self._by_orcid:
            del self._by_orcid[old]
        if new is not None:
            self._by_orcid[new] = pid

    def _add_name(self, pid: str, name: Name, during_build: bool = False) -> None:
        """Add a name for a person to the index.

        Will be called automatically from Person; do not call manually.
        """
        if not (during_build or self.is_data_loaded):
            return
        name_list = self._by_name[name]
        if pid in name_list:
            return
        if name_list:
            # Merging is transitive, so it's enough to merge with the last one
            self._similar.merge(pid, name_list[-1])
        name_list.append(pid)
        if is_verified_person_id(pid):
            verified_id_set = self._slugs_to_verified_ids[name.slugify()]
            if pid not in verified_id_set:
                for other_id in verified_id_set:
                    # Merging is transitive, so it's enough to merge with the last one
                    self._similar.merge(pid, other_id)
                    break
                verified_id_set.add(pid)

    def _remove_name(self, pid: str, name: Name) -> None:
        """Remove a name for a person from the index.

        Will be called automatically from Person; do not call manually.
        """
        if not self.is_data_loaded:
            return
        try:
            self._by_name[name].remove(pid)
            if is_verified_person_id(pid):
                self._slugs_to_verified_ids[name.slugify()].remove(pid)
        except KeyError:
            pass

    def ingest_namespec(self, name_spec: NameSpecification) -> NameSpecification:
        """Update a name specification for ingestion, potentially filling in the ID field.

        If the name specification contains an ORCID but doesn't have an ID yet, this will find the person with this ORCID and fill in their ID; if it doesn't exist yet, it will create a new person with a "verified" ID and fill in the new, generated ID.  The supplied name specification will be modified in-place, but also returned.

        Parameters:
            name_spec: The name specification on the paper, volume, etc.

        Returns:
            The name specification as it should be used for the new ingestion material.
        """
        if name_spec.orcid is None or name_spec.id is not None:
            return name_spec

        if (person := self.get_by_orcid(name_spec.orcid)) is not None:
            name_spec.id = person.id
            # Make sure the name used here is listed for this person
            person.add_name(name_spec.name)
        else:
            # Need to create a new person; generate name slug for the ID
            pid = name_spec.name.slugify()
            if pid in self.data:
                # ID is already in use; add last four digits of ORCID to disambiguate
                pid = f"{pid}-{name_spec.orcid[-4:].lower()}"

            self.add_person(
                Person(
                    id=pid,
                    parent=self.parent,
                    names=[name_spec.name] + name_spec.variants,
                    orcid=name_spec.orcid,
                    is_explicit=True,
                )
            )
            name_spec.id = pid

        return name_spec

    def resolve_namespec(
        self, name_spec: NameSpecification, allow_creation: bool = False
    ) -> Person:
        """Resolve a name specification to a person, potentially creating a new unverified person instance.

        Parameters:
            name_spec: The name specification on the paper, volume, etc.
            allow_creation: If True, will instantiate a new Person object with an unverified ID if no person matching `name_spec` exists.  Defaults to False.

        Returns:
            The person represented by `name_spec`.  If `name_spec.id` is set, this will determine the person to resolve to.  Otherwise, the slugified name will be used to find a matching person; an explicitly-defined (verified) person can be returned if exactly one such person exists and does not have `disable_name_matching` set.  In all other cases, it will resolve to an unverified person.

        Raises:
            NameSpecResolutionError: If `name_spec` cannot be resolved to a Person and `allow_creation` is False.
            PersonDefinitionError: If `name_spec.id` is set, but either the ID or the name used with the ID has not been defined in `people.yaml`. (Inherits from NameSpecResolutionError)
        """
        name = name_spec.name
        if (pid := name_spec.id) is not None:
            # Explicit ID given – should be explicitly defined in people.yaml
            if pid not in self.data or not (person := self.data[pid]).is_explicit:
                raise PersonDefinitionError(
                    name_spec, f"ID '{pid}' wasn't defined in people.yaml"
                )
            if not person.has_name(name):
                raise PersonDefinitionError(
                    name_spec,
                    f"ID '{pid}' was used with name '{name}' that wasn't defined in people.yaml",
                )
            if name_spec.orcid is not None and name_spec.orcid != person.orcid:
                raise PersonDefinitionError(
                    name_spec,
                    f"ID '{pid}' was used with ORCID '{name_spec.orcid}', but people.yaml has '{person.orcid}'",
                )
        else:
            # No explicit ID given
            if name_spec.orcid is not None:
                exc1 = NameSpecResolutionError(
                    name_spec,
                    "NameSpecification defines an ORCID without an ID",
                )
                exc1.add_note(
                    "To specify an ORCID on a paper, the person needs to have an entry in `people.yaml` and be used with an explicit ID."
                )
                raise exc1

            # Generate slug for name matching
            slug = name.slugify()

            # Check if the slugified name matches any verified IDs
            matching_ids = list(self._slugs_to_verified_ids.get(slug, []))
            if (
                len(matching_ids) == 1
                and not (person := self.data[matching_ids[0]]).disable_name_matching
            ):
                # Slug unambiguously maps to person and name matching not disabled
                pid = person.id
                if not person.has_name(name):
                    person.add_name(name, inferred=True)
                    self._add_name(pid, name, during_build=True)

            else:
                # Resolve to unverified ID
                pid = UNVERIFIED_PID_FORMAT.format(pid=slug)

                if pid in self.data:
                    # Unverified ID already exists; assume it's the same person
                    person = self.data[pid]
                    if not person.has_name(name):
                        # If the name scores higher than the current canonical
                        # one, we also assume we should set this as the
                        # canonical one
                        if name.score() > person.canonical_name.score():
                            person._set_canonical_name(name, inferred=True)
                        else:
                            person.add_name(name, inferred=True)
                        self._add_name(pid, name, during_build=True)
                elif allow_creation:
                    # Unverified ID doesn't exist yet; create it
                    person = Person(
                        id=pid, parent=self.parent, names=[(name, NameLink.INFERRED)]
                    )
                    self.add_person(person)
                else:
                    raise NameSpecResolutionError(
                        name_spec,
                        f"NameSpecification resolved to ID '{pid}' which doesn't exist",
                    )

        # Make sure that name variants specified here are registered
        for name in name_spec.variants:
            if not person.has_name(name):
                person.add_name(name, inferred=True)
            if name not in self._by_name:
                self._by_name[name].append(pid)
        return person

    def _add_to_index(
        self,
        namespecs: Iterable[NameSpecification],
        item_id: AnthologyIDTuple,
        during_build: bool = False,
    ) -> None:
        """Add persons to the index.

        This function should not be called manually.  It exists for internal use when registering new volumes or papers.  It encapsulates the resolution of all namespecs on an item, both to reduce repetition and to enable checks, e.g. that an item does not have two namespecs that resolve to the same person (which would indicate a logical error that we can't easily catch elsewhere).

        Arguments:
            namespecs: The NameSpecifications to register.
            item_id: The item to register for the NameSpecifications.
            during_build: If True, we are calling this during build and should not expect the index to be fully loaded yet.
        """
        if not (during_build or self.is_data_loaded):
            return

        seen_ids = set()
        for namespec in namespecs:
            person = self.resolve_namespec(namespec, allow_creation=True)
            person.item_ids.append(item_id)
            if person.id in seen_ids:
                message = f"More than one NameSpecification resolves to '{person.id}' on the same item ({item_id})"
                if person.is_explicit:
                    raise NameSpecResolutionError(namespec, message)
                else:
                    warnings.warn(NameSpecResolutionWarning(namespec, message))
            seen_ids.add(person.id)

    def save(self, path: Optional[StrPath] = None) -> None:
        """Save the `people.yaml` file.

        Arguments:
            path: The filename to save to. If None, defaults to the parent Anthology's `people.yaml` file.
        """
        if path is None:  # pragma: no cover
            self.parent._warn_if_in_default_path()
            path = self.path

        data = {}
        for person in self.values():
            if not person.is_explicit:
                continue

            attrib: dict[str, Any] = {
                "names": [
                    _YAMLName(name)
                    for (name, link_type) in person._names
                    if link_type == NameLink.EXPLICIT
                ],
                "comment": person.comment,
                "degree": person.degree,
                "disable_name_matching": person.disable_name_matching,
                "orcid": person.orcid,
                "similar": person.similar_ids,
            }
            data[person.id] = {k: v for k, v in attrib.items() if v}

        with open(path, "w", encoding="utf-8") as f:
            yaml.dump(data, f, allow_unicode=True, Dumper=Dumper)



# Copyright 2023-2024 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for convenient logging."""

import logging
from rich.console import Console
from rich.logging import RichHandler
from typing import cast, Any, Optional
import warnings
from ..config import config, primary_console


def get_logger() -> logging.Logger:
    """
    Returns:
        A library-specific logger instance.
    """
    return logging.getLogger(config.get("logger_name"))


class SeverityTracker(logging.Handler):
    """Tracks the highest log-level that was sent to the logger.

    If this class is added as a log handler, it can be used to check if any errors or exceptions were logged.

    Attributes:
        highest (int): The highest log-level that was sent to the logger.
    """

    def __init__(self, level: int = logging.NOTSET) -> None:
        super().__init__(level=level)
        self.highest: int = logging.NOTSET

    def emit(self, record: logging.LogRecord) -> None:
        if record.levelno > self.highest:
            self.highest = record.levelno


def setup_rich_logging(
    console: Optional[Console] = None, **kwargs: object
) -> SeverityTracker:
    """Set up a logger that uses rich markup and severity tracking.

    This function is intended to be called in a script. It calls [logging.basicConfig][] and is therefore not executed by default, as applications may wish to setup their loggers differently.

    Parameters:
        console: A [rich.Console][] object to output log messages to.
        **kwargs: Any keyword argument will be forwarded to [logging.basicConfig][].  If logging handlers are defined here, they will be preserved in addition to the handlers added by this function.  If 'level' is not given, it will default to 'INFO'.

    Returns:
        The severity tracker, so that it can be used to check the highest emitted log level.
    """
    if console is None:
        console = primary_console
    log_config: dict[str, object] = dict(
        level=kwargs.pop("level", logging.INFO),
        format="%(message)s",
        datefmt="[%X]",
        handlers=[],
    )
    log_config.update(kwargs)
    tracker = SeverityTracker()
    cast(list[logging.Handler], log_config["handlers"]).extend(
        [RichHandler(console=console), tracker]
    )
    logging.basicConfig(**log_config)  # type: ignore

    def showwarning(
        message: object, category: type[Any], *args: Any, **kwargs: Any
    ) -> None:
        logging.warning(f"{category.__name__}: {message}", stacklevel=3)

    warnings.showwarning = showwarning

    return tracker



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class EventLink(Enum):
    """How a volume ID was connected to an Event."""

    EXPLICIT = "explicit"
    """Volume ID is explicitly listed in <colocated> block in XML."""

    INFERRED = "inferred"
    """Volume ID was inferred to belong to Event through venue association."""


class PaperDeletionType(Enum):
    """Type of deletion of a paper."""

    RETRACTED = "retracted"
    """Paper was retracted.  A retraction occurs when serious, unrecoverable errors are discovered, which drastically affect the findings of the original work."""

    REMOVED = "removed"
    """Paper was removed.  A removal occurs in rare circumstances where serious ethical or legal issues arise, such as plagiarism."""


class PaperType(Enum):
    """Type of paper.

    Currently only exists to support a few paper instances that are marked up as 'backmatter'.
    """

    PAPER = "paper"
    """A regular paper."""

    FRONTMATTER = "frontmatter"
    """The frontmatter of a volume."""

    BACKMATTER = "backmatter"
    """The backmatter of a volume."""


class VolumeType(Enum):
    """Type of publication a volume represents."""

    JOURNAL = "journal"
    """A journal issue."""

    PROCEEDINGS = "proceedings"
    """A conference/workshop proceedings volume."""



# Copyright 2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions providing converters and validators for [attrs](https://www.attrs.org/) classes."""

from __future__ import annotations

import attrs
from attrs import validators
import datetime
from typing import Any, Callable, Optional, TypeVar, TYPE_CHECKING
import re

from .ids import AnthologyIDTuple

if TYPE_CHECKING:
    from ..collections import Paper, Volume, Event


RE_WRAPPED_TYPE = re.compile(r"^([^\[]*)\[(.*)\]$")
T = TypeVar("T")


def track_modifications(
    obj: Paper | Volume | Event, attr: attrs.Attribute[Any], value: T
) -> T:
    if attr.name != "parent":
        obj.collection.is_modified = True
    return value


def validate_anthology_id_tuple(cls: Any, attr: attrs.Attribute[Any], value: Any) -> None:
    """Validate that an AnthologyIDTuple has the correct format."""
    if (
        isinstance(value, tuple)
        and len(value) == 3
        and isinstance(value[0], str)
        and isinstance(value[1], (type(None), str))
        and isinstance(value[2], (type(None), str))
    ):
        return

    raise TypeError(
        f"'{attr.name}' must be AnthologyIDTuple (got {value!r})",
        attr,
        AnthologyIDTuple,
        value,
    )


def auto_validate_types(
    cls: type, fields: list[attrs.Attribute[Any]]
) -> list[attrs.Attribute[Any]]:
    """Add validators to attrs classes based on their type annotations.

    Intended to be used with the `field_transformer` parameter of [`@attrs.define`][attrs.define].

    Supported type annotations:
      - str, int
      - `FileReference` and derived classes
      - `Name`, `NameSpecification`
      - `MarkupText`
      - `Optional[<type>]`
      - `list[<type>]`
      - `tuple[<type>, ...]`

    The purpose of this function is to reduce the need for explicitly adding validators to the classes in `acl_anthology.collections` and `acl_anthology.people.person`.  It does _not_ automatically validate classes _defined_ in `acl_anthology.collections`, as that would lead to circular imports.

    See also: <https://www.attrs.org/en/stable/extending.html#transform-fields>
    """
    from ..files import (
        FileReference,
        PDFReference,
        PDFThumbnailReference,
        AttachmentReference,
        EventFileReference,
        VideoReference,
    )
    from ..people import Name, NameSpecification
    from ..text import MarkupText

    known_types = {
        t.__name__: t
        for t in {
            str,
            int,
            FileReference,
            PDFReference,
            PDFThumbnailReference,
            AttachmentReference,
            EventFileReference,
            VideoReference,
            MarkupText,
            Name,
            NameSpecification,
        }
    }

    def make_validator(field_type: Any) -> Optional[Callable[..., Any]]:
        if not isinstance(field_type, str):
            return None

        # Handle Optional[], list[], etc.
        if (m := RE_WRAPPED_TYPE.match(field_type)) is not None:
            match m.group(1):
                case "Optional":
                    if (inner := make_validator(m.group(2))) is not None:
                        return validators.optional(inner)
                case "dict":
                    dict_parts = m.group(2).split(", ")
                    if (
                        len(dict_parts) == 2
                        and (key_inner := make_validator(dict_parts[0])) is not None
                        and (value_inner := make_validator(dict_parts[1])) is not None
                    ):
                        return validators.deep_mapping(
                            key_validator=key_inner, value_validator=value_inner
                        )
                case "list":
                    if (inner := make_validator(m.group(2))) is not None:
                        return validators.deep_iterable(
                            member_validator=inner,
                            iterable_validator=validators.instance_of(list),
                        )
                case "tuple":
                    tuple_parts = m.group(2).split(", ")
                    # Only tuples of variable length with a single type
                    if (
                        len(tuple_parts) == 2
                        and tuple_parts[1] == "..."
                        and (inner := make_validator(tuple_parts[0])) is not None
                    ):
                        return validators.deep_iterable(
                            member_validator=inner,
                            iterable_validator=validators.instance_of(tuple),
                        )
            # unsupported
            return None

        # Handle known types
        if (type_ := known_types.get(field_type)) is not None:
            return validators.instance_of(type_)
        elif field_type == "AnthologyIDTuple":
            return validate_anthology_id_tuple

        # unsupported
        return None

    results = []
    for field in fields:
        # Don't modify field if validator already defined
        if field.validator is not None:
            results.append(field)
            continue

        if (validator := make_validator(field.type)) is not None:
            results.append(field.evolve(validator=validator))
        else:
            results.append(field)

    return results


def int_to_str(value: Any) -> Any:
    """Convert an int to str, and leave unchanged otherwise.

    Intended to be used as a converter for [attrs.field][].
    """
    if isinstance(value, int):
        return str(value)
    return value


def date_to_str(value: Any) -> Any:
    """Convert a [date][datetime.date] or [datetime][datetime.datetime] object to str (in ISO format), and leave unchanged otherwise.

    Intended to be used as a converter for [attrs.field][].
    """
    if isinstance(value, datetime.date):
        return value.isoformat()
    elif isinstance(value, datetime.datetime):
        return value.date().isoformat()
    return value



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from attrs import define, field
from typing import TYPE_CHECKING

from ..containers import SlottedDict
from ..exceptions import AnthologyDuplicateIDError, AnthologyInvalidIDError
from .bibkeys import BibkeyIndex
from .collection import Collection

if TYPE_CHECKING:
    from ..anthology import Anthology


@define
class CollectionIndex(SlottedDict[Collection]):
    """Index object through which all collections can be accessed.

    Provides dictionary-like functionality mapping collection IDs to [Collection][acl_anthology.collections.collection.Collection] objects in the Anthology.

    Attributes:
        parent: The parent Anthology instance to which this index belongs.
        bibkeys: A [BibkeyIndex](acl_anthology.collections.bibkeys.BibkeyIndex] instance.
        is_data_loaded: A flag indicating whether the XML directory has already been indexed.
    """

    parent: Anthology = field(repr=False, eq=False)
    bibkeys: BibkeyIndex = field(init=False, repr=False, eq=False)
    is_data_loaded: bool = field(init=False, repr=True, default=False)

    def __attrs_post_init__(self) -> None:
        self.bibkeys = BibkeyIndex(self)

    def load(self) -> None:
        """Find all XML data files and index them by their collection ID.

        Note:
            Currently assumes that XML files are **always** named according to the collection ID they
            contain; i.e., a file named "L16.xml" *must* contain the collection with ID "L16".
        """
        if self.is_data_loaded:
            return
        # Implementation note: The sorting here is mainly there for ensuring a
        # consistent order of volumes when building the website, e.g. in the
        # volume list for venues.
        for xmlpath in sorted(self.parent.datadir.glob("xml/*.xml")):
            # Assumes that XML files are **always** named as their collection
            # IDs.  --- Alternatively, could peek at the first two lines of the
            # file to parse only the <collection id="..."> tag?
            collection_id = xmlpath.name[:-4]
            self.data[collection_id] = Collection(collection_id, self, xmlpath)
        self.is_data_loaded = True

    def create(self, id: str) -> Collection:
        """Create a new [Collection][acl_anthology.collections.collection.Collection] object.

        Parameters:
            id: The ID of the new collection.

        Returns:
            The created [Collection][acl_anthology.collections.collection.Collection] object.

        Raises:
            AnthologyDuplicateIDError: If a collection with the given ID already exists.
            AnthologyInvalidIDError: If the given ID is not well-formed.
        """
        if not self.is_data_loaded:
            self.load()  # required to check for ID clashes
        if not id[0].isdigit():
            raise AnthologyInvalidIDError(id, "Collection ID must start with a digit")
        if id in self.data:
            raise AnthologyDuplicateIDError(id, "Collection ID already exists")

        collection = Collection(
            id=id,
            parent=self,
            path=self.parent.datadir / "xml" / f"{id}.xml",
        )
        collection.is_data_loaded = True
        collection.is_modified = True
        self.data[id] = collection

        return collection



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions implementing conversions to and from LaTeX/BibTeX formats."""

from __future__ import annotations

import re
from functools import lru_cache
from lxml import etree
from typing import cast, Optional, TypeAlias, TYPE_CHECKING

if TYPE_CHECKING:
    from ..people.name import NameSpecification
    from ..text import MarkupText

    SerializableAsBibTeX: TypeAlias = None | str | MarkupText | list[NameSpecification]
    """Any type that can be supplied to `make_bibtex_entry`."""

from .logging import get_logger
from .xml import append_text

from pylatexenc.latexencode import (
    UnicodeToLatexEncoder,
    UnicodeToLatexConversionRule,
    RULE_DICT,
)
from pylatexenc.latexwalker import (
    LatexWalker,
    LatexNode,
    LatexCharsNode,
    LatexGroupNode,
    LatexMacroNode,
    LatexMathNode,
    LatexSpecialsNode,
    get_default_latex_context_db as get_default_latexwalker_context_db,
)
from pylatexenc.macrospec import MacroSpec, SpecialsSpec
from pylatexenc.latex2text import (
    LatexNodes2Text,
    MacroTextSpec,
    SpecialsTextSpec,
    get_default_latex_context_db as get_default_latex2text_context_db,
)

log = get_logger()

################################################################################
### UNICODE TO LATEX (BIBTEX)
################################################################################

LATEXENC = UnicodeToLatexEncoder(
    conversion_rules=[
        UnicodeToLatexConversionRule(
            RULE_DICT,
            {
                ord("‘"): "`",  # defaults to \textquoteleft
                ord("’"): "'",  # defaults to \textquoteright
                ord("“"): "``",  # defaults to \textquotedblleft
                ord("”"): "''",  # defaults to \textquotedblright
                ord("–"): "--",  # defaults to \textendash
                ord("—"): "---",  # defaults to \textemdash
                ord("í"): "\\'i",  # defaults to using dotless \i
                ord("ì"): "\\`i",
                ord("î"): "\\^i",
                ord("ï"): '\\"i',
            },
        ),
        "defaults",
    ],
    replacement_latex_protection="braces-all",
    unknown_char_policy="keep",
    unknown_char_warning=False,
)
"""A UnicodeToLatexEncoder instance intended for BibTeX generation."""

BIBTEX_FIELD_NEEDS_ENCODING = {"journal", "address", "publisher", "note"}
"""Any BibTeX field whose value should be LaTeX-encoded first."""

BIBTEX_MONTHS = {
    "january": "jan",
    "february": "feb",
    "march": "mar",
    "april": "apr",
    "may": "may",
    "june": "jun",
    "july": "jul",
    "august": "aug",
    "september": "sep",
    "october": "oct",
    "november": "nov",
    "december": "dec",
}
"""A mapping of month names to BibTeX macros."""

RE_OPENING_QUOTE_DOUBLE = re.compile(
    r"""
     (\A|(?<=\s))    # must be start of the string or come after whitespace
     ({``}|{''}|'')  # match double apostrophe, optionally in braces
     (?!}|\s)        # must not come before whitespace or closing brace }
     """,
    re.X,
)
RE_OPENING_QUOTE_SINGLE = re.compile(
    r"""
     (\A|(?<=\s))  # must be start of the string or come after whitespace
     ({`}|{'}|')   # match single apostrophe, optionally in braces
     (?!'|}|\s)    # must not come before whitespace, closing brace, or another apostrophe
     """,
    re.X,
)
RE_CLOSING_QUOTE_DOUBLE = re.compile(
    r"""
     (?<!\\)       # must not come after backslash
     {''}          # match double apostrophe in braces
     (?=\W|\Z)     # must be end of the string or come before a non-word character
     """,
    re.X,
)
RE_CLOSING_QUOTE_SINGLE = re.compile(
    r"""
     (?<!\\)       # must not come after backslash
     {'}           # match single apostrophe in braces
     (?=\W|\Z)     # must be end of the string or come before a non-word character
     """,
    re.X,
)

RE_HYPHENS_BETWEEN_NUMBERS = re.compile(r"(?<=[0-9])(-|–|—)(?=[0-9])")


def bibtex_convert_month(spec: str) -> str:
    """Converts a month string to BibTeX macros.

    Arguments:
        spec: A month specification, as stored in the metadata.

    Returns:
        A BibTeX macro corresponding to the month specification, if possible. If the string contains digits or is otherwise not parseable, it is returned unchanged with quotes around it.
    """
    text = spec.lower()
    if text in BIBTEX_MONTHS:  # most common case; map e.g. march -> mar
        return BIBTEX_MONTHS[text]
    if text in BIBTEX_MONTHS.values():  # already a month spec
        return text
    # Find embedded month strings
    text = f'"{text}"'
    for month, macro in BIBTEX_MONTHS.items():
        if month in text:
            text = text.replace(month, f'" # {macro} # "')
            text = " # ".join(filter(lambda k: k != '""', text.split(" # ")))
    return text


def has_unbalanced_braces(string: str) -> bool:
    """Checks if a string has unbalanced curly braces."""
    c = 0
    for char in string:
        if char == "{":
            c += 1
        elif char == "}":
            c -= 1
        if c < 0:
            return True
    return c != 0


@lru_cache
def latex_encode(text: Optional[str]) -> str:
    """
    Arguments:
        text: A string that does *not* contain any LaTeX commands.

    Returns:
        The input string encoded for use in LaTeX/BibTeX.
    """
    if text is None:
        return ""
    text = cast(str, LATEXENC.unicode_to_latex(text))
    return text


def latex_convert_quotes(text: str) -> str:
    """
    Arguments:
        text: An arbitrary string.

    Returns:
        The input string with LaTeX quotes converted into proper opening and closing quotes, removing braces around them, if necessary.

    Note:
        This is called during the conversion from our XML markup to LaTeX. Straight quotation marks (`"`) will have been converted to double apostrophes, usually in braces (`{''}`), by pylatexenc; this function applies regexes to turn them into appropriate opening/closing quotes with the braces removed.

    Examples:
        >>> latex_convert_quotes("This {''}great{''} example")
        "This ``great'' example"
    """
    text = RE_OPENING_QUOTE_DOUBLE.sub("``", text)
    text = RE_OPENING_QUOTE_SINGLE.sub("`", text)
    text = RE_CLOSING_QUOTE_DOUBLE.sub("''", text)
    text = RE_CLOSING_QUOTE_SINGLE.sub("'", text)
    return text


def make_bibtex_entry(
    bibtype: str, bibkey: str, fields: list[tuple[str, SerializableAsBibTeX]]
) -> str:
    """Turn a list of field/value pairs into a BibTeX entry.

    Values will be LaTeX-formatted if necessary, and can also be empty, in which case they are automatically omitted.

    Arguments:
        bibtype: The BibTeX type for the entry.
        bibkey: The BibTeX key for the entry.
        fields: A list of tuples of the form (key, value) specifying the fields to include in the entry.

    Returns:
        A fully formatted BibTeX entry.
    """
    from ..people.name import NameSpecification
    from ..text import MarkupText

    lines = [f"@{bibtype}{{{bibkey},"]
    for key, value in fields:
        if not value:
            continue
        if isinstance(value, MarkupText):
            value = value.as_latex()
        elif isinstance(value, list) and isinstance(value[0], NameSpecification):
            value = namespecs_to_bibtex(value)
        elif isinstance(value, str):
            if key in BIBTEX_FIELD_NEEDS_ENCODING:
                value = latex_encode(value)
            elif key == "month":
                value = bibtex_convert_month(value)
            elif key == "pages":
                value = RE_HYPHENS_BETWEEN_NUMBERS.sub("--", value)
        else:
            raise TypeError(f"Unsupported type for BibTeX field: {type(value)}")
        if has_unbalanced_braces(value):
            raise ValueError(f"BibTeX field '{key}' has unbalanced braces: {value}")

        # Quote (if necessary) and append
        if key == "month":
            quoted = value
        else:
            # Make sure not to use "" to quote values when they contain "
            quoted = f'"{value}"' if '"' not in value else f"{{{value}}}"
        lines.append(f"    {key} = {quoted},")
    lines[-1] = lines[-1][:-1]  # cut off last comma
    lines.append("}")
    return "\n".join(lines)


def namespecs_to_bibtex(namespecs: list[NameSpecification]) -> str:
    """Convert a list of NameSpecifications to a BibTeX-formatted entry.

    Arguments:
        namespecs: A list of names to be included in the BibTeX entry.

    Returns:
        A BibTeX-formatted string representing the given names.
    """
    return "  and\n      ".join(spec.name.as_bibtex() for spec in namespecs)


################################################################################
### LATEX TO UNICODE/XML
################################################################################

# The logic implemented here is largely based on our old
# `bin/latex_to_unicode.py` which was mainly authored by David Chiang.

LATEX_MACRO_TO_XMLTAG = {
    "emph": "i",
    "em": "i",
    "textit": "i",
    "it": "i",
    "textsl": "i",
    "sl": "i",
    "textbf": "b",
    "bf": "b",
    "url": "url",
}
"""A mapping of LaTeX macros to Anthology XML tags."""

LATEX_CITE_MACROS = {"cite", "citep", "citet", "newcite", "citeauthor", "citeyear"}
"""A set of LaTeX macros that will be treated as citation macros."""

LW_CONTEXT = get_default_latexwalker_context_db()
LW_CONTEXT.add_context_category(
    "unhandled special characters",
    prepend=True,
    macros=[
        MacroSpec("textcommabelow", "{"),
    ],
    specials=[
        SpecialsSpec("`"),
        SpecialsSpec("'"),
    ],
)
LW_CONTEXT.add_context_category(
    "common macros",
    prepend=True,
    macros=[
        MacroSpec("href", "{{"),
    ],
)

L2T_CONTEXT = get_default_latex2text_context_db()
L2T_CONTEXT.add_context_category(
    "citations",
    prepend=True,
    macros=[
        MacroTextSpec(macro, simplify_repl=r"(CITATION)") for macro in LATEX_CITE_MACROS
    ],
)
L2T_CONTEXT.add_context_category(
    "unhandled special characters",
    prepend=True,
    macros=[
        MacroTextSpec("textcommabelow", simplify_repl="%s\N{COMBINING COMMA BELOW}"),
        MacroTextSpec("hwithstroke", simplify_repl="ħ"),
        MacroTextSpec("Hwithstroke", simplify_repl="Ħ"),
    ],
    specials=[
        SpecialsTextSpec("`", "‘"),
        SpecialsTextSpec("'", "’"),
    ],
)
L2T_CONTEXT.add_context_category(
    "common macros",
    prepend=True,
    macros=[
        # \href: drop the URL but keep the text – we could append the URL in a
        # <url> tag, but this cannot be done here, we would have to add this to
        # _parse_nodelist_to_element as another special case
        MacroTextSpec("href", simplify_repl=r"%(2)s")
    ],
)
LATEX_TO_TEXT = LatexNodes2Text(strict_latex_spaces=True, latex_context=L2T_CONTEXT)


def _is_trivial_math(node: LatexMathNode) -> bool:
    """Helper function to determine whether or not a LatexMathNode contains only 'trivial' content that doesn't require a <tex-math> node.

    Currently, a math node is considered 'trivial' if it only contains numbers, spaces, and a few allowed characters (e.g. commas, dots, percentage signs).
    """
    content = node.latex_verbatim().strip("$").replace(r"\%", "%")
    return all(c.isspace() or c.isdigit() or c in (".,@%~") for c in content)


def _should_wrap_in_fixed_case(node: LatexGroupNode) -> bool:
    """Helper function to determine whether or not a LatexGroupNode should produce a <fixed-case> tag."""
    if len(node.nodelist) == 0 or node.delimiters != ("{", "}"):
        return False
    if node.latex_verbatim().startswith("{\\"):
        # {\...} does *not* protect case
        return False
    if node.nodelist[0].isNodeType(LatexMathNode):
        # Don't mark {$...$}
        return False
    if node.nodelist[0].isNodeType(LatexSpecialsNode):
        # Don't mark {``}, {--}, etc.
        return False
    return True


def _parse_nodelist_to_element(
    nodelist: list[LatexNode],
    element: etree._Element,
    use_fixed_case: bool,
    in_macro: bool = False,
) -> None:
    """Parse a list of LaTeX nodes into an XML element using the Anthology markup format.

    Arguments:
        nodelist: The list of parsed LaTeX nodes.
        element: An XML element into which the parsed nodes will be added.
        use_fixed_case: Flag indicating whether <fixed-case> protection should be applied.
        in_macro: Flag indicating whether this function was called by recursing into a macro node. (Do not set this manually.)

    Returns:
        None; the XML element is modified in-place.
    """
    for node in nodelist:
        if node is None:
            continue  # pragma: no cover
        elif node.isNodeType(LatexCharsNode):
            # Plain text
            append_text(element, node.chars)
        elif node.isNodeType(LatexMacroNode):
            # LaTeX macro
            if (tag := LATEX_MACRO_TO_XMLTAG.get(node.macroname)) is not None:
                # This macro should get its own XML tag (e.g. \textbf -> <b>)
                subelem = etree.SubElement(element, tag)
                subnodes = node.nodeargd.argnlist
                _parse_nodelist_to_element(
                    subnodes, subelem, use_fixed_case, in_macro=True
                )
            elif node.macroname in LATEX_CITE_MACROS:
                # A citation command such as \cite{...}
                append_text(element, LATEX_TO_TEXT.macro_node_to_text(node))
            elif node.macroname == "\\":
                # Special case: explicit linebreak \\
                append_text(element, "\n")
            else:
                # This macro either represents a special characters, such as
                # \v{c} or \"I, or is some other macro we do not explicitly
                # handle; in both cases, we fall back on Latex2Text’s default
                # conversion, which can be influenced by L2T_CONTEXT
                append_text(element, LATEX_TO_TEXT.macro_node_to_text(node))
        elif node.isNodeType(LatexGroupNode):
            # Bracketed group, such as {...} or [...]
            if not in_macro and _should_wrap_in_fixed_case(node):
                # Protect this with <fixed-case>, then recurse
                subelem = etree.SubElement(element, "fixed-case")
                _parse_nodelist_to_element(node.nodelist, subelem, False)
            elif node.delimiters == ("{", "}"):
                # Just recurse
                _parse_nodelist_to_element(node.nodelist, element, use_fixed_case)
            else:
                # Skip [...] or <...> groups
                pass
        elif node.isNodeType(LatexMathNode):
            # Math node
            if _is_trivial_math(node):
                # Just append as text
                append_text(element, LATEX_TO_TEXT.math_node_to_text(node))
            else:
                # Keep verbatim, but wrap in <tex-math>
                subelem = etree.SubElement(element, "tex-math")
                subelem.text = node.latex_verbatim().strip("$")
        elif node.isNodeType(LatexSpecialsNode):
            # TODO: Is this always the correct way?
            append_text(element, LATEX_TO_TEXT.specials_node_to_text(node))
        else:
            # Comments or environments
            log.warning(f"Unhandled node type: {node.nodeType}")


def parse_latex_to_xml(
    latex_input: str, use_fixed_case: bool = True, use_heuristics: bool = False
) -> etree._Element:
    """Convert a string with LaTeX markup into the Anthology XML format.

    Arguments:
        latex_input: A string potentially including LaTeX markup.
        use_fixed_case: Flag indicating whether <fixed-case> protection should be applied.
        use_heuristics: If True, will apply some heuristics to determine if certain symbols should be interpreted as plain text rather than LaTeX; e.g., it will prevent percentage signs from being interpreted as LaTeX comments.  Set this to True when dealing with inputs that could either be plain text or LaTeX.

    Returns:
        An XML element representing the given LaTeX input in the Anthology XML format for markup strings.

    Note:
        This is a potentially lossy conversion, as the Anthology XML format only represents a small subset of LaTeX commands.  Unhandled commands will be dropped, but emit a warning in the logger.
    """
    if use_heuristics:
        # % is probably percent (not a comment delimiter)
        latex_input = re.sub(r"(?<!\\)%", r"\%", latex_input)

        # Use a heuristic to decide whether ~ means "approximately" or is a tie
        latex_input = re.sub(r"(?<=[ (])~(?=\d)", r"\\textasciitilde ", latex_input)
        latex_input = re.sub(r"^~(?=\d)", r"\\textasciitilde ", latex_input)

    element = etree.Element("root")
    walker = LatexWalker(latex_input, latex_context=LW_CONTEXT)
    nodelist, *_ = walker.get_latex_nodes()
    _parse_nodelist_to_element(nodelist, element, use_fixed_case)
    return element
